import numpy
from larcv import larcv

l2 = numpy.array([[1.0,2.0,3.0], [4.0,5.0,6.0], [7.0,8.0,9.0], [3.0, 5.0, 0.0]],dtype="float32")

print 'Input Matrix (numpy array)'
print l2

k=larcv.as_image2d(l2)
print k.meta().rows(),k.meta().cols()
print
print '1D array rep. of Image2D'
for x in k.as_vector(): print x,
print
print
print 'row/col representation of Image2D'
for r in xrange(k.meta().rows()):
    for c in xrange(k.meta().cols()):
        print k.pixel(r,c),
    print
        
ar = larcv.as_ndarray(k)
print ar



import pytest
import unittest
import random
import uuid

import larcv
from larcv import queueloader,  data_generator

from collections import OrderedDict

queue_io_sparsetensor3d_cfg_template = '''
{name}: {{
  Verbosity:       2
  EnableFilter:    false
  RandomAccess:    0 
  RandomSeed:      0
  InputFiles:      [{input_files}]
  ProcessType:     ["BatchFillerSparseTensor3D"]
  ProcessName:     ["test_{name}"]

  ProcessList: {{
    test_{name}: {{
      Tensor3DProducer: "{producer}"
      MaxVoxels: 100
      UnfilledVoxelValue: -999
    }}
  }}
}}
'''


def create_sparsetensor3d_file(file_name, rand_num_events, n_projections=1):
    voxel_set_list = data_generator.build_sparse_tensor(rand_num_events, n_projections = n_projections)
    data_generator.write_sparse_tensors(file_name, voxel_set_list, dimension=3, n_projections=n_projections)




@pytest.mark.parametrize('make_copy', [True, False])
@pytest.mark.parametrize('batch_size', [2])
def test_sparsetensor3d_queueio(tmpdir, make_copy, batch_size, n_reads=10):


    queueio_name = "queueio_{}".format(uuid.uuid4())


    # First, create a file to write to:
    file_name = str(tmpdir + "/test_queueio_sparsetensor3d_{}.h5".format(queueio_name))

    # Next, write some sparsetensor3ds to that file:
    create_sparsetensor3d_file(file_name, rand_num_events=25)


    # Generate a config for this 
    config_contents = queue_io_sparsetensor3d_cfg_template.format(
        name        = queueio_name,
        input_files = file_name,
        producer    = "test",
        )

    config_file = tmpdir + "/test_queueio_sparsetensor3d_{}.cfg".format(queueio_name) 

    with open(str(config_file), 'w') as _f:
        _f.write(config_contents)

    # Prepare data managers:
    io_config = {
        'filler_name' : queueio_name,
        'filler_cfg'  : str(config_file),
        'verbosity'   : 3,
        'make_copy'   : make_copy
    }

    data_keys = OrderedDict({
        'label': 'test_{}'.format(queueio_name), 
        })



    li = queueloader.queue_interface()
    li.prepare_manager('primary', io_config, batch_size, data_keys)


    for i in range(n_reads):
        data = li.fetch_minibatch_data('primary')
        assert(data['label'].shape[0] == batch_size)


@pytest.mark.distributed_test
@pytest.mark.parametrize('make_copy', [True, False])
@pytest.mark.parametrize('local_batch_size', [2])
def test_sparsetensor3d_queueio_distributed(tmpdir, make_copy, local_batch_size, n_reads=10):

    from larcv import distributed_queue_interface
    from mpi4py import MPI
    comm = MPI.COMM_WORLD

    root_rank = comm.Get_size() - 1
    comm_size = comm.Get_size()

    queueio_name = "queueio_{}".format(uuid.uuid4())

    config_file = tmpdir + "/test_queueio_sparsetensor3d_{}.cfg".format(queueio_name) 

    # Only generate file content on one rank:
    if comm.Get_rank() == 0:
        # First, create a file to write to:
        file_name = str(tmpdir + "/test_queueio_sparsetensor3d_{}.h5".format(queueio_name))
    
        # Next, write some sparsetensor3ds to that file:
        create_sparsetensor3d_file(file_name, rand_num_events=25)

        # Generate a config for this 
        config_contents = queue_io_sparsetensor3d_cfg_template.format(
            name        = queueio_name,
            input_files = file_name,
            producer    = "test", #"sbndvoxels",
            )

        with open(str(config_file), 'w') as _f:
            _f.write(config_contents)

    # Put a barrier so that the config file and all contents are present:
    comm.Barrier()

    # Broadcast the config file and name so that all ranks know which one to use:
    config_file = comm.bcast(config_file, root=0)
    queueio_name = comm.bcast(queueio_name, root=0)

    # Prepare data managers:
    io_config = {
        'filler_name' : queueio_name,
        'filler_cfg'  : str(config_file),
        'verbosity'   : 3,
        'make_copy'   : make_copy
    }

    data_keys = OrderedDict({
        'label': 'test_{}'.format(queueio_name), 
        })

    li = distributed_queue_interface.queue_interface()

    # Scale the number of images to the mpi comm size:
    li.prepare_manager('primary', io_config, comm_size * local_batch_size, data_keys, color=0)

    for i in range(n_reads):
        data = li.fetch_minibatch_data('primary')
        li.prepare_next('primary')
        bs = local_batch_size
        assert(data['label'].shape[0] == bs)


if __name__ == "__main__":
    # test_sparsetensor3d_queueio("./", make_copy=False, batch_size=2,  n_reads=10)
    # test_sparsetensor3d_queueio("./", make_copy=False, batch_size=2,  n_reads=10)
    test_sparsetensor3d_queueio_distributed("./", make_copy=False, local_batch_size=2, n_reads=10)
    print("Success")



import pytest
import unittest
import larcv

from larcv import data_generator

from random import Random
random = Random()

@pytest.fixture()
def rand_num_events():
    return random.randint(1, 5)


@pytest.mark.parametrize('dimension', [1, 2, 3])
@pytest.mark.parametrize('n_projections', [1,2,3])
def test_write_tensor(tmpdir, rand_num_events, dimension, n_projections):

    event_image_list = data_generator.build_tensor(rand_num_events, n_projections=n_projections, dimension=dimension)

    random_file_name = str(tmpdir + "/test_write_tensor.h5")

    data_generator.write_tensor(random_file_name, event_image_list, dimension)


@pytest.mark.parametrize('dimension', [1, 2, 3])
@pytest.mark.parametrize('n_projections', [1,2,3])
def test_write_read_tensor(tmpdir, rand_num_events, dimension, n_projections):

    import numpy

    random_file_name = str(tmpdir + "/test_write_read_tensor.h5")

    event_image_list = data_generator.build_tensor(rand_num_events, n_projections=n_projections, dimension=dimension)

    data_generator.write_tensor(random_file_name, event_image_list, dimension)

    read_event_image_list = data_generator.read_tensor(random_file_name, dimension)

    # Check the same number of events came back:
    assert(len(read_event_image_list) == rand_num_events)
    for event in range(rand_num_events):
        # Check the same number of projections per event:
        assert(len(read_event_image_list[event]) == len(event_image_list[event]))
        for projection in range(n_projections):
            # Check the same number of voxels:
            input_image = event_image_list[event][projection]
            read_image = read_event_image_list[event][projection]

            assert(numpy.sum(input_image) - numpy.sum(read_image) < 1e-3)
            
            assert(numpy.sum(numpy.fabs(input_image)) - numpy.sum(numpy.fabs(read_image)) < 1e-3)

            assert(input_image.shape == read_image.shape)

            assert(numpy.mean(input_image - read_image) < 1e-3)

            
            break
        break





if __name__ == '__main__':
    tmpdir = "./"
    rand_num_events = 3
    dimension = 1
    n_projections = 3
    # test_write_tensor(tmpdir, rand_num_events, dimension, n_projections)
    test_write_read_tensor(tmpdir, rand_num_events, dimension, n_projections)




#!/usr/bin/env python
import sys,os,argparse
from larcv import larcv

# This script is a simplification of run_processordb but without the proddb dependence.
# It incorporates command line arguments but runs stand-alone with just larcv.

parser = argparse.ArgumentParser(description='LArCV ProcessDriver execution script (use proddb)')

parser.add_argument('-c','--config',
                    type=str, dest='cfg',
                    help='string, Config file',required=True)

parser.add_argument('-il','--input-larcv',required=True,
                    dest='larcv_fin',nargs='+',
                    help='string or list, Input larcv file name[s] (Required)')

parser.add_argument('-nevents','--num-events',
                    type=int, dest='nevents', default=0,
                    help='integer, Number of events to process')

parser.add_argument('-nskip','--num-skip',
                    type=int, dest='nskip', default=0,
                    help='integer, Number of events to skip')

parser.add_argument('-ol','--output-larcv',default='',
                    type=str, dest='larcv_fout',
                    help='string,  Output larcv file name (optional)')

parser.add_argument('-oa','--output-ana',default='',
                    type=str, dest='ana_fout',
                    help='string,  Output analysis file name (optional)')

args = parser.parse_args()


if len(args.larcv_fin) == 0:
    print('No input files found')
    sys.exit(1)


proc = larcv.ProcessDriver('ProcessDriver')

proc.configure(args.cfg)

if args.larcv_fout != '':
    proc.override_output_file(args.larcv_fout)

proc.override_ana_file(args.ana_fout)

flist = larcv.VectorOfString()
for f in args.larcv_fin:
    print(f)
    flist.push_back(f)

proc.override_input_file(flist)

proc.initialize()

proc.batch_process(args.nskip,args.nevents)

print("Number of entries processed: " + str(proc.io().get_n_entries_out()))
print("Output file name: " + str(proc.io().get_file_out_name()))


proc.finalize()



import unittest

import larcv

from random import Random
random = Random()


# Testing import of Point.h objects:
def test_import_Point_h():
    pt = larcv.Point2D()
    pt = larcv.Point3D()

def test_import_BBox_h():
    bb = larcv.BBox2D()
    bb = larcv.BBox3D()


# Testing import of Vertex.h objects:
def test_import_Vertex_h():
    vert1 = larcv.Vertex()
    vert2 = larcv.Vertex(1,2,3,4)
    assert(vert1 != vert2)
    assert(vert1 < vert2)
    vert2.reset()
    assert(vert1 == vert2)

# Testing import of Particle.h objects:
def test_import_Particle_h():
    part1 = larcv.Particle()

# Testing import of Voxel.h objects:
def test_import_Voxel_h_Voxel():

    v = larcv.Voxel()

def test_import_Voxel_h_VoxelSet():

    vs = larcv.VoxelSet()

def test_import_Voxel_h_VoxelSetArray():

    vsa = larcv.VoxelSetArray()
    
def test_import_ImageMeta_h():

    im = larcv.ImageMeta2D()
    im = larcv.ImageMeta3D()

def test_import_EventID_h():

    im = larcv.EventID()

# Event Base is abstract and therefore should NOT import:
def test_import_EventBase_h():

    try:
        eb = larcv.EventBase()
        assert(False)
    except:
        assert(True)


# There is no need to wrap data product factory, so no test:
# def test_import_DataProductFactory_h():
#     # dfb = larcv.DataProductFactoryBase()
#     df = larcv.DataProductFactory()

def test_import_IOManager_h():
    io = larcv.IOManager()

def test_import_EventParticle_h():
    ep = larcv.EventParticle()


# def test_vector_double():
#     vec = larcv.VectorOfDouble()
#     vec.push_back(random.uniform(-1e5, 1e5))  

# def test_vector_sizet():
#     vec = larcv.VectorOfSizet()
#     vec.push_back(random.randint(1, 2e4))  




import unittest
import numpy
import pytest

from random import Random
random = Random()

N_CHECKS = 1

@pytest.mark.parametrize('dimension', [2,3])
def test_import(dimension):
    import larcv
    if dimension == 2:
        pt = larcv.Point2D()
    elif dimension == 3:
        pt = larcv.Point3D()
    else:
        assert False



import pytest
import larcv
import numpy

from larcv import data_generator

# Testing Voxel objects:
def test_Voxel_h_Voxel():
    vox1 = larcv.Voxel()
    vox2 = larcv.Voxel(0, 0.0)
    vox3 = larcv.Voxel(10,50.0)

    assert(vox1.id() == larcv.kINVALID_VOXELID)
    assert(vox2.id() == 0)
    assert(vox3.id() == 10)


    assert(vox1.value() == larcv.kINVALID_FLOAT)
    assert(vox2.value() == 0.0)
    assert(vox3.value() == 50.)

def test_Voxel_h_VoxelSet():

    vs = larcv.VoxelSet()
    
    n_voxels = 10
    for i in range(n_voxels):
        vs.emplace(i, i, False)

    vec = vs.as_vector()

    assert(len(vec) == n_voxels) 
    vs.clear_data()

    n_voxels = 11
    for j in range(n_voxels):
        vs.insert(larcv.Voxel(i+j, i+j))

    vs.clear_data()
    n_voxels = 12
    for j in range(n_voxels):
        vs.add(larcv.Voxel(i+j, i+j))


    assert(vs.size() == n_voxels) 


def test_Voxel_h_VoxelSetArray():

    vsa = larcv.VoxelSetArray()
    
    n_voxel_sets = 3
    for i in range(n_voxel_sets):
        n_voxels = 10*(i+1)
        vs = larcv.VoxelSet()
        vs.id(i)
        for j in range(n_voxels):
            vs.emplace(j, j, False)

        vsa.insert(vs)

    assert(vsa.size() == n_voxel_sets)
    assert(vsa.sum() > 0)


def image_meta_factory(dimension):
    if dimension == 1:
        return larcv.ImageMeta1D()
    elif dimension == 2:
        return larcv.ImageMeta2D()
    elif dimension == 3:
        return larcv.ImageMeta3D()
    elif dimension == 4:
        return larcv.ImageMeta4D()
    else:
        raise Expection("Can't do anything with dimension == ", dimension)


@pytest.mark.parametrize('dimension', [2,3])
@pytest.mark.parametrize('pooling', [larcv.kPoolAverage, larcv.kPoolMax, larcv.kPoolSum])
def test_sparse_tensor_downsample(dimension, pooling):


    # Create image Meta:
    meta = image_meta_factory(dimension)
    meta.set_projection_id(0)
    for dim in range(dimension):
        L = 10.
        N = 128
        meta.set_dimension(dim, L, N)
    
    if dimension == 2:
        st = larcv.SparseTensor2D()
    if dimension == 3:
        st = larcv.SparseTensor3D()
    st.meta(meta)
    

    # Get a set of voxels:
    voxel_set_list = data_generator.build_sparse_tensor(1, n_projections = 1)
    indexes = voxel_set_list[0][0]['indexes']
    values = voxel_set_list[0][0]['values']
    n_voxels = voxel_set_list[0][0]['n_voxels']
    for j in range(n_voxels):
        if pooling == larcv.kPoolMax:
            # Only use positive values for max pooling.  
            # Negative values have different behavior in sparse vs dense
            # max pooling.
            st.emplace(larcv.Voxel(indexes[j], numpy.abs(values[j])), False)
        else:
            st.emplace(larcv.Voxel(indexes[j], values[j]), False)

    # Dense downsampling is tested against skimage
    # Here, test against dense downsampling

    compression = 2

    st_dense = st.to_tensor()
    st_dense_compressed = st_dense.compress(compression, pooling).as_array()

    st_compressed = st.compress(compression, pooling)

    st_compressed_dense = st_compressed.dense()
    print(st_dense.as_array())

    # Do some checks:
    assert numpy.abs( (st_compressed_dense.sum() - st_dense_compressed.sum()) / st_dense_compressed.sum() )< 1e-6
    max_index = numpy.prod(st_compressed_dense.shape)
    for i in range(50):
        index = numpy.random.randint(0, max_index)
        assert numpy.abs(st_compressed_dense.take(index) - st_dense_compressed.take(index) )< 1e-4







import pytest
import unittest
import random
import uuid

import larcv
from larcv import queueloader,  data_generator

from collections import OrderedDict

queue_io_sparsetensor2d_cfg_template = '''
{name}: {{
  Verbosity:       2
  EnableFilter:    false
  RandomAccess:    0 
  RandomSeed:      0
  InputFiles:      [{input_files}]
  ProcessType:     ["BatchFillerSparseTensor2D"]
  ProcessName:     ["test_{name}"]

  ProcessList: {{
    test_{name}: {{
      Tensor2DProducer: "{producer}"
      MaxVoxels: 100
      UnfilledVoxelValue: -999
      Channels: {channels}
    }}
  }}
}}
'''


def create_sparsetensor2d_file(file_name, rand_num_events, n_projections):
    voxel_set_list = data_generator.build_sparse_tensor(rand_num_events, n_projections = n_projections)
    data_generator.write_sparse_tensors(file_name, voxel_set_list, dimension=2, n_projections=n_projections)

@pytest.mark.parametrize('make_copy', [True, False])
@pytest.mark.parametrize('batch_size', [2])
@pytest.mark.parametrize('n_projections', [1,2])
def test_sparsetensor2d_queueio(tmpdir, make_copy, batch_size, n_projections, n_reads=10):


    queueio_name = "queueio_{}".format(uuid.uuid4())


    # First, create a file to write to:
    file_name = str(tmpdir + "/test_queueio_sparsetensor2d_{}.h5".format(queueio_name))

    # Next, write some sparsetensor2ds to that file:
    create_sparsetensor2d_file(file_name, rand_num_events=25, n_projections=n_projections)


    # Generate a config for this 
    channels = list(range(n_projections))
    config_contents = queue_io_sparsetensor2d_cfg_template.format(
        name        = queueio_name,
        input_files = file_name,
        producer    = "test",
        channels    = channels,
        )

    config_file = tmpdir + "/test_queueio_sparsetensor2d_{}.cfg".format(queueio_name) 

    with open(str(config_file), 'w') as _f:
        _f.write(config_contents)

    # Prepare data managers:
    io_config = {
        'filler_name' : queueio_name,
        'filler_cfg'  : str(config_file),
        'verbosity'   : 3,
        'make_copy'   : make_copy
    }

    data_keys = OrderedDict({
        'label': 'test_{}'.format(queueio_name), 
        })



    li = queueloader.queue_interface()
    li.prepare_manager('primary', io_config, batch_size, data_keys)


    for i in range(n_reads):
        data = li.fetch_minibatch_data('primary')
        assert(data['label'].shape[0] == batch_size)


if __name__ == "__main__":
    test_sparsetensor2d_queueio("./", make_copy=False, batch_size=2, n_projections=1, n_reads=10)
    test_sparsetensor2d_queueio("./", make_copy=False, batch_size=2, n_projections=2, n_reads=10)
    print("Success")



import sys
import os
import time
import copy
from collections import OrderedDict
import socket, zlib
import random

import numpy
from mpi4py import MPI

from . queueloader    import larcv_queueio
from . larcv_io_enums import ReadOption, RandomAccess

class queue_interface(object):

    def __init__(self,
            verbose             = False,
            random_access_mode  = "random_blocks",
            entry_comm          = MPI.COMM_WORLD,
            io_comm             = MPI.COMM_WORLD,
            seed                = None ):

        '''init function

        Not much to store here, just a dict of dataloaders and the keys to access their data.

        Queue loaders are manually triggered IO, not always running, so
        '''
        object.__init__(self)
        self._queueloaders  = {}
        self._data_keys     = {}
        self._dims          = {}
        self._verbose       = verbose
        self._random_access = RandomAccess[random_access_mode]
        self._minibatch_size = {}

        self._entry_comm = entry_comm
        self._io_comm    = io_comm

        if seed is not None:
            random.seed(seed)
            numpy.random.seed(seed)

        self._count = {}
        self._warning = True

        self._queue_prev_entries = {}
        self._queue_next_entries = {}

    def no_warnings(self):
        self._warning = False

    def get_next_batch_indexes(self, mode, minibatch_size):

        # Using the random_access parameter, determine which entries to read:

        # In this function, we promote the "next" to "prev".  Do that first:
        if mode not in self._queue_next_entries:
            self._queue_next_entries[mode] = None
        self._queue_prev_entries[mode] = self._queue_next_entries[mode]
        # (If this has not been run, they will both be set to None)

        if self._random_access == RandomAccess.serial_access:
            # Figure out the last batch's highest entry:
            if mode in self._queue_prev_entries:
                if self._queue_prev_entries[mode] is not None:
                    last_entry = numpy.max(self._queue_prev_entries[mode])
                else:
                    last_entry = -1
            else:
                last_entry = -1

            next_last_entry = minibatch_size + last_entry + 1

            n_entries = self._queueloaders[mode].fetch_n_entries()

            if next_last_entry < n_entries:
                next_entries = numpy.arange(minibatch_size, dtype=numpy.int32) + last_entry + 1
            else:
                # Create an array to cover the entries till the last one ...
                next_entries_a = numpy.arange(n_entries - 1 - last_entry, dtype=numpy.int32) + last_entry + 1
                # ... and one array for the leftover entries, starting back from zero
                next_entries_b = numpy.arange((last_entry + 1 + minibatch_size) % n_entries, dtype=numpy.int32)
                # Finally concatenate the two arrays
                next_entries = numpy.concatenate((next_entries_a, next_entries_b))

        elif self._random_access == RandomAccess.random_blocks:
            # How many entries are there?
            n_entries = self._queueloaders[mode].fetch_n_entries()
            # The number of available choices is the number of entries - minibatch_size - 1:
            n_choices = n_entries - minibatch_size - 1
            start_entry = numpy.random.randint(low=0, high=n_choices, size=1)
            next_entries = numpy.arange(minibatch_size) + start_entry

        else:  # self._random_access == RandomAccess.random_events
            # Choose randomly, but require unique indexes:
            n_entries = self._queueloaders[mode].fetch_n_entries()
            next_entries = random.sample(range(n_entries), minibatch_size)

        next_entries = numpy.asarray(next_entries, dtype=numpy.int32)
        self._queue_next_entries[mode] = next_entries


        return next_entries

    def coordinate_next_batch_indexes(self, mode, comm, root_rank = 0):
        '''
        This function is a little naieve (sp?).  But it will take the root rank,
        use it to determine the next batch indexes, then scatter those indexes.

        It will scatter it evenly to the entire comm that is passed in.
        '''

        comm_size = comm.Get_size()

        if comm.Get_rank() == root_rank:
            set_entries = self.get_next_batch_indexes(mode, self._minibatch_size[mode])


        # Create a buffer for the data to scatter
        # Which is just a reference to the numpy array
        sendbuff = None
        if comm.Get_rank() == root_rank:
            sendbuff = set_entries

        local_size = int(self._minibatch_size[mode] / comm_size)

        if (self._minibatch_size[mode] % comm_size is not 0):
            print ('You have requested to scatter {} image(s) over {} ranks. This is not possible as the number are not divisible.'.format(self._minibatch_size[mode], comm_size))
            raise Exception("Please change either the minibatch size or the number or ranks to scatter to.")

        # The recvbuff must be properly sized:
        recvbuff = numpy.empty((local_size), dtype=numpy.int32)

        # Scatterv will scatter numpy arrays, and note the automatic lookup of
        # dtypes from numpy to MPI.  If you are getting crazy, undefined or NaN values,
        # the dtype is a good start for investigating.
        comm.Scatter(sendbuff,
                     recvbuff,
                     root=root_rank)

        #print (self._rank, self._local_rank, 'done Scatterv!')



        return recvbuff


    def prepare_manager(self, mode, io_config, minibatch_size, data_keys, color):
        '''Prepare a manager for io

        Creates an instance of larcv_threadio for a particular file to read.

        Arguments:
            mode {str} -- The mode of training to store this threadio under (typically "train" or "TEST" or similar)
            io_config {dict} -- the io config dictionary.  Required keys are: 'filler_name', 'verbosity', and 'filler_cfg'
            data_keys_override {dict} -- If desired, you can override the keys for dataacces,

        Raises:
            Exception -- [description]
        '''


        if mode in self._queueloaders:
            raise Exception("Can not prepare manager for mode {}, already exists".format(mode))

        # Check that the required keys are in the io config:
        for req in ['filler_name', 'verbosity', 'filler_cfg']:
            if req not in io_config:
                raise Exception("io_config for mode {} is missing required key {}".format(mode, req))


        start = time.time()

        # Initialize and configure a manager:
        io = larcv_queueio()
        io.configure(io_config, color)
        self._queueloaders.update({mode : io})
        self._minibatch_size[mode] = minibatch_size

        # Queue loaders are manually triggered IO, not always running, so
        # there is no "start_manager" function.  Everything is manual.
        # First, tell it what the entries for the first batch to read:
        self.prepare_next(mode)

        while self._queueloaders[mode].is_reading():
            # print(self._queueloaders[mode].is_reading())
            time.sleep(0.01)

        # Then, we promote those entries to the "current" batch:
        io.pop_current_data()
        io.next(store_entries=True,store_event_ids=True)
        self.prepare_next(mode)

        # Note that there is no "next" data pipelined yet.

        # Store the keys for accessing this datatype:
        self._data_keys[mode] = data_keys

        # Read and save the dimensions of the data:
        self._dims[mode] = {}
        for key in self._data_keys[mode]:
            self._dims[mode][key] = self._queueloaders[mode].fetch_data(self._data_keys[mode][key]).dim()


        end = time.time()

        # Print out how long it took to start IO:
        if self._verbose:
            sys.stdout.write("Time to start {0} IO: {1:.2}s\n".format(mode, end - start))


        return

    def prepare_next(self, mode, set_entries = None):
        '''Set in motion the processing of the next batch of data.

        Triggers the queue loader to start reading the next set of data


        '''
        # Which events should we read?
        if set_entries is None:
            set_entries = self.coordinate_next_batch_indexes(mode, comm=self._entry_comm)
            # set_entries = self.get_next_batch_indexes(mode, self._minibatch_size[mode])

        self._queueloaders[mode].set_next_batch(set_entries)
        self._queueloaders[mode].prepare_next()

        self._count[mode] = 0

        # t.daemon = True
        return
        # return threading.Thread(target=self._queueloaders[mode].batch_process).start()

        # self._queueloaders[mode].next(store_event_ids=True, store_entries=True)

        # return

    def fetch_minibatch_data(self, mode, pop=False, fetch_meta_data=False, data_shape=None, channels="last"):
        # Return a dictionary object with keys 'image', 'label', and others as needed
        # self._queueloaders['train'].fetch_data(keyword_label).dim() as an example

        if self._count[mode] != 0:
            if self._warning:
                print("Calling fetch_minibatch_data without calling prepare_next. This will not give new data.")
                print("To quiet this wanring, call prepare_next before fetch_minibatch_data or call queueloader.no_warnings()")


        if pop:
            # This function will pop the data
            while self._queueloaders[mode].is_reading():
                time.sleep(0.001)
            self._queueloaders[mode].pop_current_data()
        else:
            if self._warning:
                print("Calling fetch_minibatch_data with pop = False.  This will give you the same data as last time.")
                print("To quiet this warning, call queueloader.no_warnings()")

        self._queueloaders[mode].next(store_entries=fetch_meta_data, store_event_ids=fetch_meta_data)
        this_data = {}

        for key in self._data_keys[mode]:
            this_data[key] = self._queueloaders[mode].fetch_data(
                self._data_keys[mode][key]).data(
                shape=data_shape, channels=channels)
            # this_data[key] = numpy.reshape(this_data[key], self._dims[mode][key])

        if fetch_meta_data:
            this_data['entries'] = self._queueloaders[mode].fetch_entries()
            this_data['event_ids'] = self._queueloaders[mode].fetch_event_ids()

        self._count[mode] += 1
        return this_data

    def fetch_minibatch_dims(self, mode):
        # Return a dictionary object with keys 'image', 'label', and others as needed
        # self._queueloaders['train'].fetch_data(keyword_label).dim() as an example
        return self._dims[mode]

    # def is_active(self):
    #     _is_active = False
    #     for mode in self._queueloaders:
    #         _is_active = _is_active and self._queueloaders[mode].

    def stop(self):

        for mode in self._queueloaders:
            while self._queueloaders[mode].is_reading():
                time.sleep(0.01)
            self._queueloaders[mode].stop_manager()

        if self._writer is not None:
            self._writer.finalize()

    def size(self, mode):
        # return the number of images in the specified mode:
        return self._queueloaders[mode].fetch_n_entries()


    def ready(self, mode):
        return self._queueloaders[mode].ready()



import pytest
import unittest
import random

import larcv

@pytest.fixture()
def tempfile(tmpdir):
    return str(tmpdir.dirpath() + "/test.h5")

@pytest.fixture()
def rand_num_events():
    return random.randint(1, 10)

def write_temp_file(tempfile, rand_num_events):


    # This function is purely to test the IO read write capabilities.
    io_manager = larcv.IOManager(larcv.IOManager.kWRITE)
    io_manager.set_out_file(tempfile)
    io_manager.initialize()
    for i in range(rand_num_events):
        io_manager.set_id(0,0,i)
        io_manager.save_entry()
    io_manager.finalize()

def read_temp_file(tempfile, rand_num_events):

    # This function is purely to test the IO read write capabilities.
    io_manager = larcv.IOManager(larcv.IOManager.kREAD)
    io_manager.add_in_file(tempfile)
    io_manager.initialize()

    return io_manager.get_n_entries()


def test_file_open(tempfile, rand_num_events):

    write_temp_file(tempfile, rand_num_events)
    n_read = read_temp_file(tempfile, rand_num_events)

    assert(n_read == rand_num_events)



import numpy

import larcv

from random import Random
random = Random()


 
def write_tensor(file_name, event_image_list, dimension): 

    io_manager = larcv.IOManager(larcv.IOManager.kWRITE)
    io_manager.set_out_file(file_name)
    io_manager.initialize()

    for event in range(len(event_image_list)):
        io_manager.set_id(1001, 0, event)
        images = event_image_list[event]

        if dimension == 1:
            ev_tensor = io_manager.get_data("tensor1d","test")
        if dimension == 2:
            ev_tensor = io_manager.get_data("image2d","test")
        if dimension == 3:
            ev_tensor = io_manager.get_data("tensor3d","test")
        if dimension == 4:
            ev_tensor = io_manager.get_data("tensor4d","test")


        for projection in range(len(images)):
            if dimension == 1: tensor = larcv.Tensor1D(images[projection])
            if dimension == 2: tensor = larcv.Tensor2D(images[projection])
            if dimension == 3: tensor = larcv.Tensor3D(images[projection])
            if dimension == 4: tensor = larcv.Tensor4D(images[projection])

            ev_tensor.append(tensor)

        io_manager.save_entry()

    io_manager.finalize()

    return

def read_tensor(file_name, dimensions):

    
    from copy import copy

    io_manager = larcv.IOManager(larcv.IOManager.kREAD)
    io_manager.add_in_file(file_name)
    io_manager.initialize()

    event_image_list = []
     
    for i in range(io_manager.get_n_entries()):
        event_image_list.append([])

        io_manager.read_entry(i)
        
        # Get a piece of data, sparse tensor:
        if dimensions == 1:
            ev_tensor = io_manager.get_data("tensor1d","test")
        if dimensions == 2:
            ev_tensor = io_manager.get_data("image2d","test")
        if dimensions == 3:
            ev_tensor = io_manager.get_data("tensor3d","test")
        if dimensions == 4:
            ev_tensor = io_manager.get_data("tensor4d","test")

        print("Number of images read: ", ev_tensor.size())
        for projection in range(ev_tensor.size()):
            image = ev_tensor.tensor(projection).as_array()
            event_image_list[i].append(copy(image))

    return event_image_list

def build_tensor(rand_num_events, n_projections, dimension=2, shape=None):
    import numpy

    event_image_list = []

    for i in range(rand_num_events):
        event_image_list.append([])

        # Get a piece of data, tensor2d:
        for projection in range(n_projections):
            if shape is None:
                shape = []
                for dim in range(dimension):
                    if dimension < 3:
                        shape.append(random.randint(1, 1e3))
                    else:
                        shape.append(random.randint(1, 20))
                    
            raw_image = numpy.random.random(shape).astype("float32")
            event_image_list[i].append(raw_image)

    return event_image_list

def write_sparse_clusters(file_name, voxel_set_array_list, dimension=2, n_projections=3):


    import copy

    io_manager = larcv.IOManager(larcv.IOManager.kWRITE)
    io_manager.set_out_file(file_name)
    io_manager.initialize()


    # For this test, the meta is pretty irrelevant as long as it is consistent
    meta_list = []
    for projection in range(n_projections):
        if dimension == 2:
            meta_list.append(larcv.ImageMeta2D())
        else:
            meta_list.append(larcv.ImageMeta3D())

        for dim in range(dimension):
            L = 10.
            N = 128
            meta_list[-1].set_dimension(dim, L, N)

        meta_list[-1].set_projection_id(projection)

    for i in range(len(voxel_set_array_list)):
        io_manager.set_id(1001, 0, i)
        # Get a piece of data, sparse tensor:
        if dimension== 2:
            ev_cluster = io_manager.get_data("cluster2d","test")
        else:
            ev_cluster = io_manager.get_data("cluster3d","test")

        # Holder for the voxels to store:


        for projection in range(n_projections):
            clusters = voxel_set_array_list[i][projection]
            if dimension == 2:
                vsa = larcv.SparseCluster2D()
            else:
                vsa = larcv.SparseCluster3D()
            for cluster in range(len(clusters)):
                vs = larcv.VoxelSet()

                vs.id(cluster)
                indexes = clusters[cluster]['indexes']
                values = clusters[cluster]['values']
                for j in range(clusters[cluster]['n_voxels']):
                    vs.emplace(indexes[j], values[j], False)
                vsa.insert(vs)          
            # sparse_cluster.set(vsa, )
            vsa.meta(meta_list[projection])
            # ev_cluster.emplace(vsa, copy.copy(meta_list[projection]))
            ev_cluster.set(vsa)

        io_manager.save_entry()





    # assert(io_manager.get_n_entries_out() == rand_num_events)

    io_manager.finalize()

    return
 
def read_sparse_clusters(file_name, dimension):



    io_manager = larcv.IOManager(larcv.IOManager.kREAD)
    io_manager.add_in_file(file_name)
    io_manager.initialize()


    voxel_set_array_list = []
     
    for event in range(io_manager.get_n_entries()):
        # append a list of projections for this event:
        voxel_set_array_list.append([])

        io_manager.read_entry(event)
        
        # Get a piece of data, sparse cluster:\
        if dimension == 2:
            ev_cluster = io_manager.get_data("cluster2d","test")
        else:
            ev_cluster = io_manager.get_data("cluster3d","test")

        for projection in range(ev_cluster.size()):
            # Append a list of clusters for this projection:
            voxel_set_array_list[event].append([])
            print("Number of clusters: ", ev_cluster.sparse_cluster(projection).size())
            sparse_cluster = ev_cluster.sparse_cluster(projection)
            print("Current voxel_set_array_list length: ", len(voxel_set_array_list))
            print("Current voxel_set_array_list[] length: ", len(voxel_set_array_list[event]))
            for cluster in range(sparse_cluster.size()):
                # Append a dict of values for this cluster
                voxel_set_array_list[event][projection].append({
                    "indexes" : [],
                    "values"  : [],
                    "n_voxels": sparse_cluster.voxel_set(cluster).size()
                    })
                for j in range(sparse_cluster.voxel_set(cluster).size()):
                    voxel_set_array_list[event][projection][cluster]['indexes'].append(sparse_cluster.voxel_set(cluster).as_vector()[j].id())
                    voxel_set_array_list[event][projection][cluster]['values'].append(sparse_cluster.voxel_set(cluster).as_vector()[j].value())



    return voxel_set_array_list

def build_sparse_cluster_list(rand_num_events, n_projections=3):


    voxel_set_array_list = []
     
    for event in range(rand_num_events):
        #  Add space for an event:
        voxel_set_array_list.append([])
        # Get a piece of data, sparse tensor:
        for projection in range(n_projections):
            # In this event, add space for a projection:
            voxel_set_array_list[event].append([])
            n_clusters = random.randint(1,25)
            for cluster in range(n_clusters):
                # print(len(voxel_set_array_list[projection][cluster]))
                n_voxels = random.randint(1,50)
                cluster_d = {                    
                    'values'  : [],
                    'indexes' : random.sample(range(128*128), n_voxels),
                    'n_voxels': n_voxels}
                for j in range(n_voxels):
                    cluster_d['values'].append(random.uniform(-1e3, 1e3) )

                voxel_set_array_list[event][projection].append(cluster_d)

    return voxel_set_array_list



def write_sparse_tensors(file_name, voxel_set_list, dimension, n_projections):


    from copy import copy
    io_manager = larcv.IOManager(larcv.IOManager.kWRITE)
    io_manager.set_out_file(file_name)
    io_manager.initialize()

    # For this test, the meta is pretty irrelevant as long as it is consistent
    meta_list = []
    for projection in range(n_projections):
        if dimension == 2:
            meta_list.append(larcv.ImageMeta2D())
        else:
            meta_list.append(larcv.ImageMeta3D())

        for dim in range(dimension):
            L = 10.
            N = 128
            meta_list[-1].set_dimension(dim, L, N)

        meta_list[-1].set_projection_id(projection)

    for i in range(len(voxel_set_list)):
        io_manager.set_id(1001, 0, i)
        
        # Get a piece of data, sparse tensor:
        if dimension== 2:
            ev_sparse = io_manager.get_data("sparse2d","test")
        else:
            ev_sparse = io_manager.get_data("sparse3d","test")


        for projection in range(n_projections):
            vs = larcv.VoxelSet()
            indexes = voxel_set_list[i][projection]['indexes']
            values = voxel_set_list[i][projection]['values']
            for j in range(voxel_set_list[i][projection]['n_voxels']):
                vs.emplace(indexes[j], values[j], False)

            ev_sparse.set(vs, meta_list[projection])
        io_manager.save_entry()


    print("Finished event loop")



    # assert(io_manager.get_n_entries_out() == rand_num_events)

    io_manager.finalize()

    return
 
def read_sparse_tensors(file_name, dimension):



    io_manager = larcv.IOManager(larcv.IOManager.kREAD)
    io_manager.add_in_file(file_name)
    io_manager.initialize()


    voxel_set_list = []
     
    for i in range(io_manager.get_n_entries()):
        voxel_set_list.append([])

        io_manager.read_entry(i)
        
        # Get a piece of data, sparse tensor:
        if dimension== 2:
            ev_sparse = io_manager.get_data("sparse2d","test")
        else:
            ev_sparse = io_manager.get_data("sparse3d","test")

        for projection in range(ev_sparse.size()):
            voxel_set_list[i].append({
                "indexes" : [],
                "values"  : [],
                "n_voxels": ev_sparse.sparse_tensor(projection).size()
                })
            for j in range(ev_sparse.sparse_tensor(projection).size()):
                voxel_set_list[i][projection]['indexes'].append(ev_sparse.sparse_tensor(projection).as_vector()[j].id())
                voxel_set_list[i][projection]['values'].append(ev_sparse.sparse_tensor(projection).as_vector()[j].value())



    return voxel_set_list

def build_sparse_tensor(rand_num_events, n_projections=3):

    voxel_set_list = []
     
    for i in range(rand_num_events):
        voxel_set_list.append([])
        # Get a piece of data, sparse tensor:
        for projection in range(n_projections):
            n_voxels = random.randint(25,50)
            voxel_set_list[i].append({
                'values'  : [],
                'indexes' : random.sample(range(128*128), n_voxels),
                'n_voxels': n_voxels
                })
            for j in range(voxel_set_list[i][projection]['n_voxels']):
                voxel_set_list[i][-1]['values'].append(random.uniform(-1e3, 1e3) )

    return voxel_set_list



def write_particles(tempfile, rand_num_events, particles_per_event=-1):

    io_manager = larcv.IOManager(larcv.IOManager.kWRITE)
    io_manager.set_out_file(tempfile)
    io_manager.initialize()


    for i in range(rand_num_events):
        io_manager.set_id(1001, 0, i)
        
        # Get a piece of data, particle:
        ev_particle = io_manager.get_data("particle","test")



        if particles_per_event == -1:
            n_particles = i + 1
        else:
            n_particles = particles_per_event

        for j in range(n_particles):
            part = larcv.Particle()
            part.energy_deposit(j)
            part.pdg_code(0)
            ev_particle.append(part)


        io_manager.save_entry()


    assert(io_manager.get_n_entries_out() == rand_num_events)

    io_manager.finalize()

    return
 
def read_particles(tempfile, use_core_driver=False):

    io_manager = larcv.IOManager(larcv.IOManager.kREAD)
    if use_core_driver: 
      io_manager.set_core_driver()
    io_manager.add_in_file(tempfile)
    io_manager.initialize()

    read_events = 0
    for i in range(io_manager.get_n_entries()):
        io_manager.read_entry(i)
        # print(io_manager.current_entry())
        event_id = io_manager.event_id()

        ev_particles = io_manager.get_data('particle', 'test')
        read_events += 1

    return read_events


import unittest



def test_import_processor_top():
    import larcv

def test_import_processor():
    import larcv

    lib = larcv.ProcessDriver()
    # lib = larcv.ProcessFactory()

def test_import_processor_base():
    import larcv

    # Processes base is virtual and should NOT import:
    try:
        lib = larcv.ProcessBase("test")
        assert(False)
    except:
        assert(True)

def test_process_driver():
    import larcv

    lib = larcv.ProcessDriver("test")
    


import unittest

# def test_logger():
#     from larcv import base
#     # logger = base.logger.get_shared()
#     for level in [base.kDEBUG, base.kINFO, base.kNORMAL, base.kWARNING, base.kERROR]:
#         base.logger.force_level(level)

#         assert(level == base.logger.get_shared().level())


from enum import Enum

class RandomAccess(Enum):
    # Read all events in order, always:
    serial_access = 0 
    # Read chunks of events in order:
    random_blocks = 1 
    # Read every event randomly from file:
    random_events = 2 

class ReadOption(Enum):
    # Use only the specified root rank to read the data, 
    # and the distribute to all other ranks:
    read_from_single_rank       = 0  

    # Use all ranks to read the data, no distribution is needed
    # This makes a copy of the file on every rank
    read_from_all_ranks_copy    = 1  

    # Use only one *local* rank to read the data, 
    # and the distribute to all other ranks in subgroup
    # This copies the file to the local rank.
    read_from_single_local_rank = 2  

    # Use every rank to read the data with MPI-IO
    # This is best for large files that are 
    # challenging to duplicate to every node.
    read_from_all_ranks_mpi     = 3



import unittest



def test_larbys():
    import larcv
    lib = larcv.larbys()
    # Config manager is deliberately excluded:
    # lib = base.ConfigManager()

def test_larbys_exception():
    import larcv
    # Creating a PSet object without a name will cause an exception.
    # We can catch that and verify we're getting a larbys exception:
    try:
        pset = larcv.PSet()
        return False
    except:
        return True


# import pytest
# import unittest
# import larcv

# from larcv import data_generator

# from random import Random
# random = Random()

# # 
# # For backward compatibility, otherwise see test_write_tensor.py
# # 

# @pytest.fixture()
# def rand_num_events():
#     return random.randint(1, 5)


# # def write_image2d(tempfile, event_image_list):

    

# #     io_manager = larcv.IOManager(larcv.IOManager.kWRITE)
# #     io_manager.set_out_file(tempfile)
# #     io_manager.initialize()


        
# #     for event in range(len(event_image_list)):
# #         io_manager.set_id(1001, 0, event)
# #         images = event_image_list[event]

# #         # Get a piece of data, image2d:
# #         ev_image2d = larcv.EventImage2D.to_image2d(io_manager.get_data("image2d","test"))

# #         for projection in range(len(images)):

# #             image2d = larcv.as_image2d(images[projection])
# #             ev_image2d.append(image2d)

# #         io_manager.save_entry()


# #     print("Finished event loop")



# #     # assert(io_manager.get_n_entries_out() == rand_num_events)

# #     io_manager.finalize()

# #     return
 
# # def read_image2d(tempfile):

    
# #     from copy import copy

# #     io_manager = larcv.IOManager(larcv.IOManager.kREAD)
# #     io_manager.add_in_file(tempfile)
# #     io_manager.initialize()


# #     event_image_list = []
     
# #     for i in range(io_manager.get_n_entries()):
# #         event_image_list.append([])

# #         io_manager.read_entry(i)
        
# #         # Get a piece of data, sparse tensor:
# #         ev_image2d = larcv.EventImage2D.to_image2d(io_manager.get_data("image2d","test"))
# #         print("Number of images read: ", ev_image2d.as_vector().size())
# #         for projection in range(ev_image2d.as_vector().size()):
# #             image = larcv.as_ndarray(ev_image2d.as_vector()[projection])
# #             event_image_list[i].append(copy(image))


# #     return event_image_list

# # def build_image2d(rand_num_events, n_projections):
# #     from larcv import larcv
# #     import numpy

# #     event_image_list = []

# #     for i in range(rand_num_events):
# #         event_image_list.append([])

# #         # Get a piece of data, image2d:
# #         for projection in range(n_projections):
# #             shape = []
# #             for dim in range(2):
# #                 shape.append(random.randint(1, 1e3))
# #             raw_image = numpy.random.random(shape).astype("float32")
# #             event_image_list[i].append(raw_image)

# #     return event_image_list


# @pytest.mark.parametrize('n_projections', [1,2,3])
# def test_write_image2d(tmpdir, rand_num_events, n_projections):

#     event_image_list = data_generator.build_image2d(rand_num_events, n_projections = n_projections)

#     random_file_name = str(tmpdir + "/test_write_image2d.h5")

#     data_generator.write_image2d(random_file_name, event_image_list)


# @pytest.mark.parametrize('n_projections', [1,2,3])
# def test_write_read_image2d(tmpdir, rand_num_events, n_projections):

#     import numpy

#     random_file_name = str(tmpdir + "/test_write_read_image2d.h5")

#     event_image_list = data_generator.build_image2d(rand_num_events, n_projections = n_projections)

#     data_generator.write_image2d(random_file_name, event_image_list)
#     read_event_image_list = data_generator.read_image2d(random_file_name)

#     # Check the same number of events came back:
#     assert(len(read_event_image_list) == rand_num_events)
#     for event in range(rand_num_events):
#         # Check the same number of projections per event:
#         assert(len(read_event_image_list[event]) == len(event_image_list[event]))
#         for projection in range(n_projections):
#             # Check the same number of voxels:
#             input_image = event_image_list[event][projection]
#             read_image = read_event_image_list[event][projection]

#             assert(numpy.sum(input_image) - numpy.sum(read_image) < 1e-3)
#             assert(numpy.sum(numpy.fabs(input_image)) - numpy.sum(numpy.fabs(read_image)) < 1e-3)

#             assert(input_image.shape == read_image.shape)

#             assert(numpy.mean(input_image - read_image) < 1e-3)





# if __name__ == '__main__':
#     tmpdir = "./"
#     rand_num_events = 10
#     n_projections = 3
#     test_write_read_image2d(tmpdir, rand_num_events, n_projections)




import unittest
import numpy
import pytest

import larcv

from random import Random
random = Random()

N_CHECKS = 1

def image_meta_factory(dimension):
    if dimension == 1:
        return larcv.ImageMeta1D()
    elif dimension == 2:
        return larcv.ImageMeta2D()
    elif dimension == 3:
        return larcv.ImageMeta3D()
    elif dimension == 4:
        return larcv.ImageMeta4D()
    else:
        raise Expection("Can't do anything with dimension == ", dimension)


@pytest.mark.parametrize('dimension', [1,2,3,4])
def test_import(dimension):
    
    try:
        im = image_meta_factory(dimension)
    except:
        assert False

@pytest.mark.parametrize('dimension', [1,2,3,4])
@pytest.mark.parametrize('execution_number', range(N_CHECKS))
def test_default_constructor(dimension, execution_number):
    
    im = image_meta_factory(dimension)

    # n_dims = random.randint(1,4)

    total_volume = 1.0
    total_voxels = 1

    _id = random.randint(1, 100)
    im.set_projection_id(_id)

    for dim in range(dimension):
        L = random.uniform(0.001, 1e4)
        N = random.randint(1, 2e4)
        O = random.uniform(0.001, 1e4)
        im.set_dimension(dim, L, N, O)
        total_volume *= L
        total_voxels *= N

    assert(im.id() == _id)
    assert(im.n_dims() == dimension)
    assert(im.total_voxels() == total_voxels)



@pytest.mark.parametrize('dimension', [1,2,3,4])
@pytest.mark.parametrize('execution_number', range(N_CHECKS))
def test_filled_constructor(dimension, execution_number):

    

    projection_id = random.randint(0,10)

    number_of_voxels = []
    image_sizes = []
    origin = []

    total_volume = 1.0
    total_voxels = 1

    for dim in range(dimension):
        L = random.uniform(0.001, 1e4)
        N = random.randint(1, 2e4)
        O = random.uniform(0.001, 1e4)
        number_of_voxels.append(N)
        image_sizes.append(L)
        origin.append(O)
        total_volume *= L
        total_voxels *= N

    if dimension == 1:
        im = larcv.ImageMeta1D(projection_id, number_of_voxels, image_sizes)
    elif dimension == 2:
        im = larcv.ImageMeta2D(projection_id, number_of_voxels, image_sizes)
    elif dimension == 3:
        im = larcv.ImageMeta3D(projection_id, number_of_voxels, image_sizes)
    elif dimension == 4:
        im = larcv.ImageMeta4D(projection_id, number_of_voxels, image_sizes)
    else:
        assert(False)

    assert(im.n_dims() == dimension)
    assert(im.total_voxels() == total_voxels)

@pytest.mark.parametrize('dimension', [1,2,3,4])
@pytest.mark.parametrize('execution_number', range(N_CHECKS))
def test_unravel(dimension, execution_number):

    im = image_meta_factory(dimension)    


    total_volume = 1.0
    total_voxels = 1

    dims = []

    for dim in range(dimension):
        L = random.uniform(0.001, 1e4)
        N = random.randint(1, 2e4)
        im.set_dimension(dim, L, N)
        total_volume *= L
        total_voxels *= N
        dims.append(N)

    # Verify that the same uniform index gets unraveled to the same value as numpy:
    for i in range(50):
        flat_index = random.randint(0, total_voxels-1)
        np_unraveled = numpy.unravel_index(flat_index, dims)
        im_unraveled = im.coordinates(flat_index)

        for d in range(dimension):
            assert(np_unraveled[d] == im_unraveled[d])

@pytest.mark.parametrize('dimension', [1,2,3,4])
@pytest.mark.parametrize('execution_number', range(N_CHECKS))
def test_ravel(dimension, execution_number):

    
    im = image_meta_factory(dimension)    

    total_volume = 1.0
    total_voxels = 1

    dims = []

    for dim in range(dimension):
        L = random.uniform(0.001, 1e4)
        N = random.randint(1, 2e4)
        im.set_dimension(dim, L, N)
        total_volume *= L
        total_voxels *= N
        dims.append(N)

    # Verify that the same coordinates get raveled to the same index as numpy:
    for i in range(50):
        
        # Create a list of indexes:
        indexes = []
        vec_of_indexes = []
        for dim in dims:
            ind = random.randint(0, dim-1)
            indexes.append(ind)
            vec_of_indexes.append(ind)

        # for 
        np_raveled = numpy.ravel_multi_index(indexes, dims)
        im_raveled = im.index(vec_of_indexes)

        for d in range(dimension):
            assert(np_raveled == im_raveled)


@pytest.mark.parametrize('dimension', [1,2,3,4])
@pytest.mark.parametrize('execution_number', range(N_CHECKS))
def test_compress(dimension, execution_number):

    
    im = image_meta_factory(dimension)    


    dims = []
    lens = []
    comp = []


    for dim in range(dimension):
        N = random.randint(1, 2e4)
        L = random.uniform(0.001, 1e4)
        im.set_dimension(dim, L, N)
        dims.append(N)
        lens.append(L)


    compression = [2]*dimension
    compressed_meta = im.compress(compression)

    # Verify that the same coordinates get raveled to the same index as numpy:
    for i in range(dimension):
        assert compressed_meta.image_size(i) == im.image_size(i) / compression[i]
        assert compressed_meta.number_of_voxels(i) == int(im.number_of_voxels(i) / compression[i])
        assert compressed_meta.origin(i) == im.origin(i) / compression[i]



import unittest


def test_import_base_full():
    import larcv
    lib = larcv.logger()
    lib = larcv.larcv_base()
    lib = larcv.PSet("test")
    lib = larcv.Watch()
    # Config manager is deliberately excluded:
    # lib = base.ConfigManager()


import larcv
import sys,time,os,signal
import numpy 

class batch_pydata(object):

   _storage_id   = -1
   _dtype        = None
   _npy_data     = None
   _dim_data     = None
   _dim_dense    = None
   _time_copy    = 0
   _time_reshape = 0
   _make_copy    = False

   def __init__(self,dtype):
      self._storage_id   = -1
      self._dtype        = dtype
      self._npy_data     = None
      self._dim_data     = None
      self._dim_dense    = None
      self._time_copy    = None
      self._time_reshape = None
      
   def batch_data_size(self):
      dsize=1
      for v in self._dim_data: dsize *= v
      return dsize

   def dtype(self): 
      return self._dtype

   def data(self, shape=None, channels="last"): 
      '''Return the data of this batch
      
      larcv is generally used to read sparse data and return in, often,
      a sparse format.  This is acheived by packing into a buffer and returning
      directly a buffer.  This function can quickly parse the buffer and expand
      to a dense or other popular sparse shape.  If shape is none, it's untouched.
      
      Keyword Arguments:
         shape {str} -- Set to "dense", "scnsparse", or "torch_geometric" to expand the data (default: {None})
         channels {str} -- If shape is "dense", channels first or last? (default: {"last"})
      
      Returns:
         [type] -- [description]
      '''
      if shape is None:
         return numpy.reshape(self._npy_data, self._dim_data)
      elif shape == "dense":
         return self.as_dense(channels)
      elif shape == "scnsparse":
         return self.as_scnsparse()
      elif shape == "torch_geometric":
         return self.as_torch_geometric()

   def dim(self):  
      return self._dim_data

   def time_copy(self): 
      return self._time_copy

   def time_reshape(self): 
      return self._time_reshape

   def set_data(self,storage_id,larcv_batchdata):
      self._storage_id = storage_id
      dim   = larcv_batchdata.dim()
      dense = larcv_batchdata.dense_dim()



      # set dimension
      # if self._dim_data is None:
      # Dimension may change from shot to shot.  Refresh every time:
      self._dim_data  = numpy.array([dim[i] for i in range(len(dim))]).astype(numpy.int32)
      self._dense_dim = numpy.array([dense[i] for i in range(len(dense))]).astype(numpy.int32)


               
      # copy data into numpy array
      ctime = time.time()
      if self._make_copy:
         if self._npy_data is None:
            # self._npy_data = numpy.array(larcv_batchdata.data())
            # print("Array type: ", self._npy_data.dtype)
            # Create an array to hold the data if it does not exits:
            self._npy_data = numpy.ndarray(shape=(larcv_batchdata.data_size()), dtype=self._dtype)
         self._npy_data = self._npy_data.reshape(self.batch_data_size())
         self._npy_data = numpy.copy(larcv_batchdata.pydata())

      else:
         self._npy_data = larcv_batchdata.pydata()



      self._time_copy = time.time() - ctime


      ctime = time.time()
      self._npy_data = numpy.reshape(self._npy_data, self._dim_data)
      # self._npy_data = self._npy_data.reshape(self._dim_data[0], int(self.batch_data_size()/self._dim_data[0]))
      self.time_data_conv = time.time() - ctime

      return

   def as_dense(self, channels="last"):

      '''
      This format converts the larcv sparse format to
      the dense format
      '''

      # If the data is from BatchFillerSparseTensor2D,
      # it will have the shape (batch_size, n_planes, n_elements_per_plane, 2 or 3)
      # where the 2 or 3 is if values are included or not.

      # If the data is from BatchFillerSparseTensor3D, it will have the
      # shape (batchsize, n_elements, 3 or 4) where the 3 or 4 
      # is if values are included or not




      if len(self._dim_data) == 4:
         # This is 2D
         # Initialize an output array.  Larcv puts channels last by default (B, W, H, Ch)
         if channels == "last":
            output_array = numpy.zeros(
               [self._dense_dim[0], self._dense_dim[2], self._dense_dim[1], self._dense_dim[3]],
               dtype=self._dtype)
         else:
            output_array = numpy.zeros(
               [self._dense_dim[0], self._dense_dim[3], self._dense_dim[2], self._dense_dim[1]],
               dtype=self._dtype)

         # Slice the input array into x, y, vals:
         x_coords = self._npy_data[:,:,:,0]
         y_coords = self._npy_data[:,:,:,1]
         val_coords = self._npy_data[:,:,:,2]
         
         # Where are the non-zero locations?
         filled_locs = val_coords != -999
         
         # Map the output locations based on non-zero locations:
         batch_index, plane_index, voxel_index = numpy.where(filled_locs)

         # Pull out only the important values:
         values  = val_coords[batch_index, plane_index, voxel_index]
         x_index = numpy.int32(x_coords[batch_index, plane_index, voxel_index])
         y_index = numpy.int32(y_coords[batch_index, plane_index, voxel_index])
         if channels == "last":
            output_array[batch_index, y_index, x_index, plane_index] = values
         else:
            output_array[batch_index, plane_index, y_index, x_index] = values
         return output_array

      elif len(self._dim_data) == 3:
         # This is 3D
         # Initialize an output array.  Larcv puts channels last by default (B, X, Y, Z, Ch==1)
         if channels == "last":
            output_array = numpy.zeros(
               [self._dense_dim[0], self._dense_dim[1], self._dense_dim[2], self._dense_dim[3], self._dense_dim[4]],
               dtype=self._dtype)
         else:
            output_array = numpy.zeros(
               [self._dense_dim[0], self._dense_dim[4], self._dense_dim[1], self._dense_dim[2], self._dense_dim[3]],
               dtype=self._dtype)
         print(output_array.shape)

         # Slice the input array into x, y, vals:
         x_coords   = self._npy_data[:,:,0]
         y_coords   = self._npy_data[:,:,1]
         z_coords   = self._npy_data[:,:,2]
         val_coords = self._npy_data[:,:,3]
         
         # Where are the non-zero locations?
         filled_locs = val_coords != -999
         
         # Map the output locations based on non-zero locations:
         batch_index, voxel_index = numpy.where(filled_locs)


         values  = val_coords[batch_index, voxel_index]
         x_index = numpy.int32(x_coords[batch_index, voxel_index])
         y_index = numpy.int32(y_coords[batch_index, voxel_index])
         z_index = numpy.int32(z_coords[batch_index, voxel_index])


         # Fill in the output tensor

         output_array[batch_index, 0, x_index, y_index, z_index] = values    

         if channels == "last":
            output_array[batch_index, x_index, y_index, z_index, 0 ] = values
         else:
            output_array[batch_index, 0, x_index, y_index, z_index ] = values
         return output_array
       



   def as_torch_geometric(self):

      '''
      This format converts the larcv sparse format to
      the tuple format required for sparseconvnet
      '''

      # If the data is from BatchFillerSparseTensor2D,
      # it will have the shape (batch_size, n_planes, n_elements_per_plane, 2 or 3)
      # where the 2 or 3 is if values are included or not.

      # If the data is from BatchFillerSparseTensor3D, it will have the
      # shape (batchsize, n_elements, 3 or 4) where the 3 or 4 
      # is if values are included or not

      if len(self._dim_data) == 4:
         # This is 2D
         return None
      elif len(self._dim_data) == 3:
         # This is 3D
         return None

   def as_scnsparse(self):

      '''
      This format converts the larcv sparse format to
      the tuple format required for sparseconvnet
      '''

      # If the data is from BatchFillerSparseTensor2D,
      # it will have the shape (batch_size, n_planes, n_elements_per_plane, 2 or 3)
      # where the 2 or 3 is if values are included or not.

      # If the data is from BatchFillerSparseTensor3D, it will have the
      # shape (batchsize, n_elements, 3 or 4) where the 3 or 4 
      # is if values are included or not

      if len(self._dim_data) == 4:
         # This is 2D
         return None
      elif len(self._dim_data) == 3:
         # This is 3D
         return None

#       # First, we can split off the features (which is the pixel value)
#       # and the indexes (which is everythin else)
#       n_dims = input_array.shape[-1]

#       split_tensors = numpy.split(input_array, n_dims, axis=-1)


#       # To map out the non_zero locations now is easy:
#       non_zero_inds = numpy.where(split_tensors[-1] != -999)

#       # The batch dimension is just the first piece of the non-zero indexes:
#       batch_size  = input_array.shape[0]
#       batch_index = non_zero_inds[0]

#       # Getting the voxel values (features) is also straightforward:
#       features = numpy.expand_dims(split_tensors[-1][non_zero_inds],axis=-1)

#       # Lastly, we need to stack up the coordinates, which we do here:
#       dimension_list = []
#       for i in range(len(split_tensors) - 1):
#         dimension_list.append(split_tensors[i][non_zero_inds])

#       # Tack on the batch index to this list for stacking:
#       dimension_list.append(batch_index)

#       # And stack this into one numpy array:
#       dimension = numpy.stack(dimension_list, axis=-1)

#       output_array = (dimension, features, batch_size,)
#       return output_array




# def larcvsparse_to_scnsparse_3d(input_array):



# def larcvsparse_to_scnsparse_2d(input_array):
#     # This format converts the larcv sparse format to
#     # the tuple format required for sparseconvnet

#     # First, we can split off the features (which is the pixel value)
#     # and the indexes (which is everythin else)

#     # To handle the multiplane networks, we have to split this into
#     # n_planes and pass it out as a list

#     n_planes = input_array.shape[1]
#     batch_size = input_array.shape[0]


#     raw_planes = numpy.split(input_array,n_planes, axis=1)

#     output_list = []
#     output_features = []
#     output_dimension = []

#     for i, plane in enumerate(raw_planes):
#         # First, squeeze off the plane dimension from this image:
#         plane = numpy.squeeze(plane, axis=1)

#         # Next, figure out the x, y, value coordinates:
#         x,y,features = numpy.split(plane, 3, axis=-1)


#         non_zero_locs = numpy.where(features != -999)
#         # Pull together the different dimensions:
#         x = x[non_zero_locs]
#         y = y[non_zero_locs]
#         p = numpy.full(x.shape, fill_value=i)
#         features = features[non_zero_locs]
#         features = numpy.expand_dims(features,axis=-1)

#         batch = non_zero_locs[0]

#         # dimension = numpy.concatenate([x,y,batch], axis=0)
#         # dimension = numpy.stack([x,y,batch], axis=-1)
#         dimension = numpy.stack([p,x,y,batch], axis=-1)

#         output_features.append(features)
#         output_dimension.append(dimension)

#     output_features = numpy.concatenate(output_features)
#     output_dimension = numpy.concatenate(output_dimension)

#     output_list = [output_dimension, output_features, batch_size]

#     return output_list






#!/usr/bin/env python
import ROOT, sys, os
for l in [x for x in os.listdir(os.environ['LARCV_LIBDIR']) if x.endswith('.so')]:
   ROOT.gSystem.Load('%s/%s' % (os.environ['LARCV_LIBDIR'],l))

from ROOT import std
from larcv import larcv

if len(sys.argv) < 2:

   print 'Usage: python',sys.argv[0],'CONFIG_FILE [LARCV_FILE1 LARCV_FILE2 ...]'
   sys.exit(1)

proc = larcv.ProcessDriver('ProcessDriver')

print "config: ",sys.argv[1]
proc.configure(str(sys.argv[1]))

if len(sys.argv) > 1:
   
   flist=ROOT.std.vector('std::string')()
   for x in xrange(len(sys.argv)-2):
      flist.push_back(sys.argv[x+2])

   proc.override_input_file(flist)

proc.initialize()

#proc.batch_process(0,3)
proc.batch_process()

proc.finalize()



from larcv import larcv
from colored_msg import colored_msg as cmsg
import os, sys
larcv.logger.force_level(0)

#
# Constants
#
MSG_LEVEL=larcv.msg.kERROR
if 'debug' in sys.argv:
    MSG_LEVEL = larcv.msg.kDEBUG
if 'info' in sys.argv:
    MSG_LEVEL = larcv.msg.kINFO

OUT_FNAME="segmask.root"
NUM_EVENT=1

ERROR_FILE_EXIST      = 1
ERROR_WRITE_INIT      = 2

if os.path.isfile(OUT_FNAME):
    cmsg.error("Test output file (%s) already exists..." % OUT_FNAME)
    sys.exit(ERROR_FILE_EXIST)

from larcv import larcv
o=larcv.IOManager(larcv.IOManager.kWRITE)
o.reset()
o.set_verbosity(MSG_LEVEL)
o.set_out_file(OUT_FNAME)

p = larcv.SegmentMask()
cfg = larcv.CreatePSetFromFile(sys.argv[1],"SegmentMask")
p.configure(cfg)
p.initialize()

if not o.initialize():
    sys.exit(ERROR_WRITE_INIT)
for idx in xrange(NUM_EVENT):

    img = larcv.Image2D(10,10)
    for x in xrange(img.as_vector().size()):
        if x%2 == 0: img.set_pixel(x,larcv.kROIEminus)
        else: img.set_pixel(x,larcv.kROIGamma)
        
    event_image1 = o.get_data(larcv.kProductImage2D,"original")
    event_image1.Append(img)
    event_image2 = o.get_data(larcv.kProductImage2D,"target")
    event_image2.Append(img)

    o.set_id(0,0,idx)
    p.process(o)
    o.save_entry()

    idx+=1

p.finalize()
o.finalize()



import sys
import os
import time

import numpy
import larcv

class larcv_writer(object):

    '''A utility class for writing data as larcv objects to file.
    
    This class doesn't reinvent any wheels, it's just meant to package everything
    into an easy to use interface for writing output files.
    '''

    def __init__(self, io_config, output_file=None):
        '''larcv_writer accesses an IOManager instance directly to manage writing
        
        In the init function, the IOManager is created and prepared for writing.
        
        Arguments:
             {} -- [description]
        '''

        print(io_config)

        self._io = larcv.IOManager(io_config)

        if output_file is not None:
            self._io.set_out_file(output_file)

        self._io.initialize()

        self._write_workers = {
            'sparse2d' : self._write_sparse2d,
            'image2d'  : self._write_image2d,
            'tensor1d' : self._write_tensor1d
        }

        pass

    def finalize(self):
        # The purpose of this function is to make sure the output file is persisted to disk:
        self._io.save_entry()
        self._io.finalize()

    def _seek(self, entry):
        '''Go to an entry
        
        Saves the current entry and seeks to a new entry.
        
        Arguments:
            entry {int} -- Desired Entry
        '''
        self._io.save_entry()
        self._io.read_entry(entry)

        return

    def _write_sparse2d(self, data, producer):
        '''Write sparse2d data to file.
        
        Data is writting under the provided producer name.  This function takes a list of data
        that it can iterate through, assigning each element of the list contains a dict with the following 
        three elements:
         - 'index' is a flattened 1D array of index values, which are the 'raveled' values
         - 'value' is the list of features
         - 'meta' is the meta data object.  It is meant to have the arugments for the constructor of ImageMeta2d,
            so it is a list in the following order: [xmin, ymin, xmax, ymax, n_x, n_y, projection_id]
        
        Arguments:
            data {list} -- list (by projection ID) of dict objects.  dict must have the keys 'index', 'value' and 'meta'
            producer {[type]} -- [description]
        '''
        _writable_data = larcv.EventSparseTensor2D.to_sparse_tensor(self._io.get_data("sparse2d", producer))


        for projection in range(len(data)):
            value = data[projection]['values']
            index = data[projection]['index']
            shape = data[projection]['shape']
            meta = larcv.ImageMeta2D()
            meta.set_projection_id(projection)
            meta.set_dimension(0, shape[0], int(shape[0]))
            meta.set_dimension(1, shape[1], int(shape[1]))

            # First, copy all of the values into a VoxelSet object:
            voxel_set = larcv.as_voxelset(value.astype(numpy.float32), index.astype(numpy.uint64))
            # Add the voxel set:
            _writable_data.set(voxel_set, meta)

        return

    def _write_image2d(self, data, producer):
        ''' Write image2d data to file

        Data is written under the provided producer name.  This function takes a list of data
        that it can iterate through, where each list element is a projection ID.
        '''


        ev_image2d = larcv.EventImage2D.to_image2d(self._io.get_data("image2d", producer))

        for projection_id, image in enumerate(data):
            meta = larcv.ImageMeta2D()
            meta.set_dimension(0, image.shape[0], image.shape[0])
            meta.set_dimension(1, image.shape[1], image.shape[1])
            meta.set_projection_id(projection_id)
            img = larcv.as_image2d_meta(image, meta)
            ev_image2d.emplace(img)
            # print(img)

        return


    def _write_tensor1d(self, data, producer):
        '''
        Write tensor1d data to file

        This can be used to save the softmax output for different categories
        Arguments:
            data {numpy} -- Data to write, a list of values
            producer {str} -- Producer key underwhich to store this data

        '''

        n_classes = len(data)

        ev_tensor = larcv.EventTensor1D.to_tensor(self._io.get_data("tensor1d", producer))

        tensor = larcv.as_tensor1d(data)

        ev_tensor.append(tensor)

        return



    def write(self, data, datatype, producer, entries, event_ids):
        '''Write data to file
        
        This function writes data to the output file.  It will check the entries and event ids
        to ensure there is not a mismatch.  You can call this function multiple times per event,
        if there are multiple things to write.  When you call it for the next event, and it has
        a new entry/event_id, it will save the previous event, seek to the next event, and continue.
        
        Arguments:
            data {numpy} -- Data to write, must meet certain formatting depending on datatype
            datatype {str} -- larcv2 datatype descriptor
            producer {srt} -- producer key underwhich to store this data
            entries {vector<int>} -- vector of file entries, size 1 for now
            event_ids {vector<larcv::EventBase>} -- vector of larcv EventBase objects, size 1 for now
        '''

        # First, check if the entries recieved for the data match the current entry
        # for the output file:

        # print("Request recieved to write at entry ", entries[0])
        # print("Current entry is ", self._io.current_entry())

        if self._io.current_entry() != entries[0]:
            self._seek(entries[0])

        if datatype not in self._write_workers:
            raise Exception("No current support to write datatypes of type " + datatype)

        else:
            self._write_workers[datatype](data, producer)


        pass




import argparse
import glob

import h5py, numpy

def main():


  parser = argparse.ArgumentParser(description='LArCV ProcessDriver execution script (use proddb)')

  parser.add_argument('-il','--input-larcv',required=True,
                      dest='larcv_fin',nargs='+',
                      help='string or list, Input larcv file name[s] (Required)')

  parser.add_argument('-ol','--output-larcv',default='',
                      type=str, dest='larcv_fout',
                      help='string,  Output larcv file name (optional)')


  args = parser.parse_args()

  file_list = args.larcv_fin
  if type(file_list) == str:
    file_list = [file_list,]

  output_file = h5py.File(args.larcv_fout, 'w')


  for f in file_list: 
    print("Adding file to output: ", f)
    input_file = h5py.File(f,'r')
    copy_file_to_output(input_file,output_file)
    input_file.close()


def copy_file_to_output(input_file, output_file):

  copy_event_id_table(input_file, output_file)

  # Make sure the data group is in the output file:
  if 'Data' not in output_file:
    output_file.create_group("Data")

  for group in input_file['Data'].keys():
    copy_data_product(input_file, output_file, group)



def copy_event_id_table(input_file, output_file):

  # First, open the table from the input file:
  input_event_id_table = input_file['Events/event_id']

  # Next, see if the table is in the output file:
  if 'Events' not in output_file:
    output_file.create_group("Events")

  if 'event_id' not in output_file["Events/"]:
    output_file.create_dataset("Events/event_id",
      data=input_event_id_table,
      # shape=input_event_id_table.dims,
      chunks=input_event_id_table.chunks,
      dtype=input_event_id_table.dtype,
      maxshape=input_event_id_table.maxshape,
      compression=input_event_id_table.compression)

  else:
    # The dataset already exists, so resize and assign it new elements:
    output_event_id_table = output_file['Events/event_id']
    original_offset = output_event_id_table.shape[0]
    new_shape = numpy.asarray(input_event_id_table.shape) + numpy.asarray(output_event_id_table.shape)
    
    output_event_id_table.resize(new_shape) 

    output_event_id_table[original_offset:] = input_event_id_table


def copy_data_product(input_file, output_file, group_name):

  # Converting a dataproduct means copying all tables from a group into
  # the same tables in the output group.  The datasets that are holding the actual 
  # data (images, voxels, particles) are all unchanged.
  # Any table that has 'extents' in it's name has to be updated.
  # In particular, the 'first' column in every extents dataset must be modified
  # by adding the total number of the last entry in the existing table (first + n)


  full_path = "Data/" + group_name + "/"


  if full_path not in output_file:
    output_file.create_group(full_path)


  for dataset in input_file[full_path].keys():


    # Get the input dataset:
    input_dataset = input_file[full_path + dataset]

    # Make sure the dataset is in the output file:
    if dataset not in output_file[full_path]:
      # print ("Creating new dataset ", dataset)
      output_file.create_dataset(full_path + dataset,
        data=input_dataset,
        # shape=input_dataset.dims,
        chunks=input_dataset.chunks,
        dtype=input_dataset.dtype,
        maxshape=input_dataset.maxshape,
        compression=input_dataset.compression)
    else:
      # print ("Appending to dataset ", dataset)
      # The output dataset already exists.  So resize it and assign new elements:
      output_table = output_file[full_path + dataset]
      # print("Existing shape: ", output_table.shape)
      original_offset = output_table.shape[0]
      new_shape = numpy.asarray(input_dataset.shape) + numpy.asarray(output_table.shape)
      # print("Input shape: ", input_dataset.shape)
      # print("New shape: ", new_shape)
    
      # print("Updated shape: ", output_table.shape)

      # Deal with the extents issue:

      if 'extents' in dataset:

        # Get the offsets for the new table before resizing:
        existing_offset = output_table['first'][-1]+ output_table['N'][-1]
        index = len(output_table)



      output_table.resize(new_shape) 

      output_table[original_offset:] = input_dataset

      # Update the offset values:
      if 'extents' in dataset:
        output_table['first', original_offset:] += existing_offset

  pass


if __name__ == "__main__":
  main()


import pytest
import unittest
import larcv

from random import Random
random = Random(1)

from larcv import data_generator

@pytest.fixture()
def tempfile(tmpdir):
    return str(tmpdir.dirpath() + "/test_sparse.h5")

@pytest.fixture()
def rand_num_events():
    return random.randint(5, 100)



# @pytest.mark.parametrize()
@pytest.mark.parametrize('dimension', [2, 3])
@pytest.mark.parametrize('n_projections', [1, 2, 3])
def test_write_sparse_clusters(tmpdir, rand_num_events, dimension, n_projections):

    voxel_set_array_list = data_generator.build_sparse_cluster_list(rand_num_events, n_projections)
    random_file_name = str(tmpdir + "/test_write_sparse_clusters.h5")
    data_generator.write_sparse_clusters(random_file_name, voxel_set_array_list, dimension, n_projections)



@pytest.mark.parametrize('dimension', [2,3])
@pytest.mark.parametrize('n_projections', [1,2,3])
def test_read_write_sparse_clusters(tmpdir, rand_num_events, dimension, n_projections):
   
    import numpy
   

    voxel_set_array_list = data_generator.build_sparse_cluster_list(rand_num_events, n_projections)
    random_file_name = str(tmpdir + "/test_write_sparse_clusters.h5")

    print(len(voxel_set_array_list))
    for e in voxel_set_array_list:
        print("  " + str(len(e)))
        for p in e:
            print("    " + str(len(p)))

    data_generator.write_sparse_clusters(random_file_name, voxel_set_array_list, dimension, n_projections)
    read_voxel_set_array_list = data_generator.read_sparse_clusters(random_file_name, dimension)

    # Check the same number of events came back:
    assert(len(read_voxel_set_array_list) == rand_num_events)
    for event in range(rand_num_events):
        # Check the same number of projections per event:
        print
        assert(len(read_voxel_set_array_list[event]) == len(voxel_set_array_list[event]))

        for projection in range(n_projections):
            # CHeck the number of clusters in this projection:
            assert(len(read_voxel_set_array_list[event][projection]) == len(voxel_set_array_list[event][projection]))

            for cluster in range(len(read_voxel_set_array_list[event][projection])):
                # Check the same number of voxels:
                input_voxelset = voxel_set_array_list[event][projection][cluster]
                read_voxelset = read_voxel_set_array_list[event][projection][cluster]
                assert(read_voxelset['n_voxels'] == input_voxelset['n_voxels'])

                # Check voxel properties:
                # Sum of indexes
                # Sum of values
                # std of values
                if input_voxelset['n_voxels'] == 0:
                    continue
                # print(input_voxelset['values'])
                assert(numpy.sum(input_voxelset['indexes']) == numpy.sum(read_voxelset['indexes']))
                assert( abs( numpy.sum(input_voxelset['values']) - numpy.sum(read_voxelset['values']) ) < 1e-3 )
                assert( abs( numpy.std(input_voxelset['values']) - numpy.std(read_voxelset['values']) ) < 1e-3 )




if __name__ == '__main__':
    tmpdir = "./"
    rand_num_events = 5
    n_projections = 3
    dimension = 2
    test_read_write_sparse_clusters(tmpdir, rand_num_events, dimension, n_projections)






from skbuild import setup  # This line replaces 'from setuptools import setup'
import argparse

import io
import os
this_directory = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Several environment variables control special build features:
if 'LARCV_WITH_MPI' in os.environ and os.environ['LARCV_WITH_MPI']:
    mpi_value='ON'
else:
    mpi_value='OFF'

if 'LARCV_WITH_OPENMP' in os.environ and os.environ['LARCV_WITH_OPENMP']:
    openmp_value='ON'
else:
    openmp_value='OFF'



setup(
    name="larcv",
    version="3.3.1",
    cmake_source_dir='src/',
    use_scm_version=True,
    include_package_data=True,
    cmake_args=[
        '-DCMAKE_OSX_DEPLOYMENT_TARGET:STRING=10.9',
        '-DCMAKE_PYTHON_BINDINGS=True',
        # '-DMPI_CXX_COMPILER={}'.format(mpicxx),
        # '-DMPI_C_COMPILER={}'.format(mpicc),
        '-DMPI:BOOL={}'.format(mpi_value),
        '-DOPENMP:BOOL={}'.format(openmp_value),
    ],
    author=['Corey Adams', 'Kazuhiro Terao', 'Taritree Wongjirad', 'Marco del Tutto'],
    author_email='corey.adams@anl.gov',
    description='C++ IO and Preprocessing package for sparse neutrino data, with H5 for IO and python bindings.',
    license='MIT',
    keywords='larcv larcv3 neutrinos hdf5 h5 deep learning IO sparse',
    project_urls={
        'Source Code': 'https://github.com/DeepLearnPhysics/larcv3'
    },
    url='https://github.com/DeepLearnPhysics/larcv3',
    scripts=['bin/merge_larcv3_files.py', 'bin/run_processor.py'],
    packages=['larcv','src/pybind11'],
    install_requires=[
        'numpy',
        'scikit-build',
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
)



import unittest
import larcv

# queue_io_sparsetensor2d_cfg_template = '''
# {name}: {{
#   Verbosity:       2
#   EnableFilter:    false
#   RandomAccess:    0 
#   RandomSeed:      0
#   InputFiles:      [{input_files}]
#   ProcessType:     ["BatchFillerSparseTensor2D"]
#   ProcessName:     ["test_{name}"]

#   ProcessList: {{
#     test_{name}: {{
#       Tensor2DProducer: "{producer}"
#       MaxVoxels: 100
#       UnfilledVoxelValue: -999
#       Channels: {channels}
#     }}
#   }}
# }}
# '''

def test_pset_creation():

  IOPSet = larcv.PSet("IOManager")
  IOPSet.add_value("Verbosity", "2")
  assert int(IOPSet["Verbosity"]) == 2

def test_pset_modification():

  IOPSet = larcv.PSet("IOManager")
  IOPSet.add_value("Verbosity", "2")
  IOPSet["Verbosity"] = str(1)
  assert int(IOPSet["Verbosity"]) == 1

def test_pset_nesting():

  overall_pset = larcv.PSet("TrainIO")

  overall_pset["Verbosity"]    = str(2)
  overall_pset["EnableFilter"] = str("false")
  overall_pset["RandomAccess"] = str(0 )
  overall_pset["RandomSeed"]   = str(0)
  overall_pset["InputFiles"]   = str([""])
  overall_pset["ProcessType"]  = str(["BatchFillerSparseTensor2D"])
  overall_pset["ProcessName"]  = str(["test_{name}"])

  process_pset = larcv.PSet("ProcessList")
  process_pset["test"] = "yes"

  overall_pset.add_pset(process_pset)




if __name__ == "__main__":
  test_pset_creation()


from larcv import larcv
from colored_msg import colored_msg as cmsg
import os, sys
larcv.logger.force_level(0)

#
# Constants
#
MSG_LEVEL=larcv.msg.kERROR
if 'debug' in sys.argv:
    MSG_LEVEL = larcv.msg.kDEBUG
if 'info' in sys.argv:
    MSG_LEVEL = larcv.msg.kINFO

OUT_FNAME="croproi.root"
NUM_EVENT=1

ERROR_FILE_EXIST      = 1
ERROR_WRITE_INIT      = 2

if os.path.isfile(OUT_FNAME):
    cmsg.error("Test output file (%s) already exists..." % OUT_FNAME)
    sys.exit(ERROR_FILE_EXIST)

from larcv import larcv
o=larcv.IOManager(larcv.IOManager.kWRITE)
o.reset()
o.set_verbosity(MSG_LEVEL)
o.set_out_file(OUT_FNAME)

p = larcv.CropROI()
cfg = larcv.CreatePSetFromFile(sys.argv[1],"CropROI")
p.configure(cfg)
p.initialize()

if not o.initialize():
    sys.exit(ERROR_WRITE_INIT)
for idx in xrange(NUM_EVENT):

    img = larcv.Image2D(10,10)
    for x in xrange(img.as_vector().size()):
        img.set_pixel(x,x%3)
        
    event_image1 = o.get_data(larcv.kProductImage2D,"original")
    event_image1.Append(img)
    event_image2 = o.get_data(larcv.kProductImage2D,"target")
    event_image2.Append(img)

    roi = larcv.ROI()
    #ImageMeta(const double width=0.,     const double height=0.,
    #          const size_t row_count=0., const size_t col_count=0,
    #          const double origin_x=0.,  const double origin_y=0.,
    #          const PlaneID_t plane=::larcv::kINVALID_PLANE)
    meta = larcv.ImageMeta(3,3,3,3,2,-2,0)
    roi.AppendBB(meta)
    event_roi = o.get_data(larcv.kProductROI,"roi")
    event_roi.Append(roi)
    o.set_id(0,0,idx)
    
    p.process(o)
    o.save_entry()

    idx+=1
p.finalize()
o.finalize()



import pytest
import unittest
import random
import larcv

from larcv import data_generator

@pytest.fixture()
def tempfile(tmpdir):
    return str(tmpdir.dirpath() + "/test_particle.h5")

@pytest.fixture()
def rand_num_events():
    return random.randint(5, 100)




def test_write_particles(tmpdir, rand_num_events):

    tempfile = str(tmpdir + "/test_write_particles.h5")

    data_generator.write_particles(tempfile, rand_num_events)
    n_read = data_generator.read_particles(tempfile)

    assert(n_read == rand_num_events)

# def test_read_particles_coredriver(tmpdir, rand_num_events):

#     tempfile = str(tmpdir + "/test_write_particles.h5")

#     write_particles(tempfile, rand_num_events)
#     n_read = read_particles(tempfile, True)
#     assert(n_read == rand_num_events)

if __name__ == '__main__':
    tmpdir = "./"
    rand_num_events = 5
    test_write_particles(tmpdir, rand_num_events)



from test_write_sparse_cluster import *

rand_num_events = 10
dimension = 3
n_projections = 3


voxel_set_list = build_sparse_cluster_list(rand_num_events, n_projections=n_projections)

tempfile = "cluster.h5"
write_sparse_clusters(tempfile, voxel_set_list, dimension, n_projections)
read_voxel_set_list = read_sparse_clusters(tempfile, dimension)



import sys,time,os,signal
import numpy
import threading
import random


import larcv
from . batch_pydata   import batch_pydata
from . larcv_io_enums import RandomAccess
from . larcv_writer   import larcv_writer

class queue_interface(object):

    def __init__(self, verbose=False, random_access_mode="random_blocks", seed=None):
        '''init function

        Not much to store here, just a dict of dataloaders and the keys to access their data.

        Queue loaders are manually triggered IO, not always running, so
        '''
        object.__init__(self)
        self._queueloaders  = {}
        self._data_keys     = {}
        self._dims          = {}
        self._verbose       = verbose
        self._random_access = RandomAccess[random_access_mode]
        self._minibatch_size = {}

        self._queue_prev_entries = {}
        self._queue_next_entries = {}
        self._count = {}

        if seed is not None:
            random.seed(seed)
            numpy.random.seed(seed)

        self._warning = True

        self._writer = None

    def no_warnings(self):
        self._warning = False

    def get_next_batch_indexes(self, mode, minibatch_size):

        # Using the random_access parameter, determine which entries to read:

        # In this function, we promote the "next" to "prev".  Do that first:
        if mode not in self._queue_next_entries:
            self._queue_next_entries[mode] = None
        self._queue_prev_entries[mode] = self._queue_next_entries[mode]
        # (If this has not been run, they will both be set to None)

        if self._random_access == RandomAccess.serial_access:
            # Figure out the last batch's highest entry:
            if mode in self._queue_prev_entries:
                if self._queue_prev_entries[mode] is not None:
                    last_entry = numpy.max(self._queue_prev_entries[mode])
                else:
                    last_entry = -1
            else:
                last_entry = -1

            next_last_entry = minibatch_size + last_entry + 1

            n_entries = self._queueloaders[mode].fetch_n_entries()

            if next_last_entry < n_entries:
                next_entries = numpy.arange(minibatch_size, dtype=numpy.int32) + last_entry + 1
            else:
                # Create an array to cover the entries till the last one ...
                next_entries_a = numpy.arange(n_entries - 1 - last_entry, dtype=numpy.int32) + last_entry + 1
                # ... and one array for the leftover entries, starting back from zero
                next_entries_b = numpy.arange((last_entry + 1 + minibatch_size) % n_entries, dtype=numpy.int32)
                # Finally concatenate the two arrays
                next_entries = numpy.concatenate((next_entries_a, next_entries_b))

        elif self._random_access == RandomAccess.random_blocks:
            # How many entries are there?
            n_entries = self._queueloaders[mode].fetch_n_entries()
            # The number of available choices is the number of entries - minibatch_size - 1:
            n_choices = n_entries - minibatch_size - 1
            start_entry = numpy.random.randint(low=0, high=n_choices, size=1)
            next_entries = numpy.arange(minibatch_size) + start_entry

        else:  # self._random_access == RandomAccess.random_events
            # Choose randomly, but require unique indexes:
            n_entries = self._queueloaders[mode].fetch_n_entries()
            next_entries = random.sample(range(n_entries), minibatch_size)


        self._queue_next_entries[mode] = next_entries

        return next_entries


    def set_next_index(self, mode, entry):
        '''
        Set the next entry to be read. Only works in serial_access mode.

        Argumements:
            mode {str} -- The mode of training to store this threadio under (typically "train" or "TEST" or similar)
            entry {int} -- The next entry to be read
        '''

        if mode not in self._queue_prev_entries:
            throw('Need to call prepare_manager first.')

        if self._random_access != RandomAccess.serial_access:
            throw('set_next_entry can only be called in serial_access mode.')

        self._queue_next_entries[mode] = [entry - 1.]
        return



    def prepare_manager(self, mode, io_config, minibatch_size, data_keys, color=None):
        '''Prepare a manager for io

        Creates an instance of larcv_threadio for a particular file to read.

        Arguments:
            mode {str} -- The mode of training to store this threadio under (typically "train" or "TEST" or similar)
            io_config {dict} -- the io config dictionary.  Required keys are: 'filler_name', 'verbosity', and 'filler_cfg'
            data_keys_override {dict} -- If desired, you can override the keys for dataacces,

        Raises:
            Exception -- [description]
        '''


        if mode in self._queueloaders:
            raise Exception("Can not prepare manager for mode {}, already exists".format(mode))

        # Check that the required keys are in the io config:
        for req in ['filler_name', 'verbosity', 'filler_cfg']:
            if req not in io_config:
                raise Exception("io_config for mode {} is missing required key {}".format(mode, req))


        start = time.time()

        # Initialize and configure a manager:
        io = larcv_queueio()
        io.configure(io_config)
        self._queueloaders.update({mode : io})
        self._minibatch_size[mode] = minibatch_size

        # Queue loaders are manually triggered IO, not always running, so
        # there is no "start_manager" function.  Everything is manual.
        # First, tell it what the entries for the first batch to read:

        # start = time.time()
        # print(self._queueloaders[mode].is_reading())
        self.prepare_next(mode)
        # print(self._queueloaders[mode].is_reading())
        # end = time.time()

        # print(end - start)
        # Then, we promote those entries to the "current" batch:
        while self._queueloaders[mode].is_reading():
            # print(self._queueloaders[mode].is_reading())
            time.sleep(0.01)

        io.pop_current_data()
        io.next(store_entries=True,store_event_ids=True)

        # Note that there is no "next" data pipelined yet.

        # Store the keys for accessing this datatype:
        self._data_keys[mode] = data_keys

        # Read and save the dimensions of the data:
        self._dims[mode] = {}
        for key in self._data_keys[mode]:
            self._dims[mode][key] = self._queueloaders[mode].fetch_data(self._data_keys[mode][key]).dim()

        end = time.time()
        self.prepare_next(mode)

        # Print out how long it took to start IO:
        if self._verbose:
            sys.stdout.write("Time to start {0} IO: {1:.2}s\n".format(mode, end - start))


        return


    def prepare_writer(self, io_config, output_file=None):

        if self._writer is not None:
            raise Exception("queue_interface doesn't yet support multiple writers.")

        # This only supports batch datasize of 1.  We can check that the reading instance
        # Has only size 1 by looking at the dims.

        key =list(self._data_keys['primary'].items())[0][0]
        if self._dims['primary'][key][0] != 1:
            raise Exception("To use the writing interface, please set batch size to 1.")

        # The writer is not an instance of queueIO but rather an instance of larcv_writer.

        # It configures a process to copy input to output and add more information as well.

        self._writer = larcv_writer(io_config, output_file)

        pass


    def prepare_next(self, mode, set_entries = None):
        '''Set in motion the processing of the next batch of data.

        Triggers the queue loader to start reading the next set of data


        '''
        # Which events should we read?
        if set_entries is None:
            set_entries = self.get_next_batch_indexes(mode, self._minibatch_size[mode])

        self._queueloaders[mode].set_next_batch(set_entries)
        self._queueloaders[mode].prepare_next()

        self._count[mode] = 0

        return
        # return threading.Thread(target=self._queueloaders[mode].batch_process).start()

        # self._queueloaders[mode].next(store_event_ids=True, store_entries=True)

        # return

    def fetch_minibatch_data(self, mode, pop=False, fetch_meta_data=False, data_shape=None, channels="last"):
        # Return a dictionary object with keys 'image', 'label', and others as needed
        # self._queueloaders['train'].fetch_data(keyword_label).dim() as an example

        if self._count[mode] != 0:
            if self._warning:
                print("Calling fetch_minibatch_data without calling prepare_next. This will not give new data.")
                print("To quiet this warning, call prepare_next before fetch_minibatch_data or call queueloader.no_warnings()")

        if pop:
            # This function will pop the data
            while self._queueloaders[mode].is_reading():
                time.sleep(0.001)
            self._queueloaders[mode].pop_current_data()
        else:
            if self._warning:
                print("Calling fetch_minibatch_data with pop = False.  This will give you the same data as last time.")
                print("To quiet this warning, call queueloader.no_warnings()")

        self._queueloaders[mode].next(store_entries=fetch_meta_data, store_event_ids=fetch_meta_data)
        this_data = {}

        self.fetch_minibatch_dims(mode)

        for key in self._data_keys[mode]:
            # self._dims[mode][key] = self._queueloaders[mode].fetch_data(self._data_keys[mode][key]).dim()
            this_data[key] = self._queueloaders[mode].fetch_data(
                self._data_keys[mode][key]).data(
                shape=data_shape, channels=channels)
            # this_data[key] = numpy.reshape(this_data[key], self._dims[mode][key])

        if fetch_meta_data:
            this_data['entries'] = self._queueloaders[mode].fetch_entries()
            this_data['event_ids'] = self._queueloaders[mode].fetch_event_ids()

        self._count[mode] += 1

        return this_data

    def fetch_minibatch_dims(self, mode):
        # Return a dictionary object with keys 'image', 'label', and others as needed
        # self._queueloaders['train'].fetch_data(keyword_label).dim() as an example
        return self._dims[mode]

    # def is_active(self):
    #     _is_active = False
    #     for mode in self._queueloaders:
    #         _is_active = _is_active and self._queueloaders[mode].

    def stop(self):

        if self._writer is not None:
            self._writer.finalize()

    def size(self, mode):
        # return the number of images in the specified mode:
        return self._queueloaders[mode].fetch_n_entries()


    def is_reading(self, mode):
        return self._queueloaders[mode].is_reading()


    def write_output(self, data, datatype, producer, entries, event_ids):
        if self._writer is None:
            raise Exception("Trying to write data with no writer configured.  Abort!")


        self._writer.write(data=data, datatype=datatype, producer=producer, entries=entries, event_ids=event_ids)

        return


class larcv_queueio (object):

    _instance_m={}

    @classmethod
    def exist(cls,name):
        name = str(name)
        return name in cls._instance_m

    @classmethod
    def instance_by_name(cls,name):
        return cls._instance_m[name]

    def __init__(self):
        self._proc = None
        self._name = ''
        self._verbose = False
        self._read_start_time = None
        self._read_end_time = None
        self._cfg_file = None
        self._storage = {}
        self._event_entries = None
        self._event_ids = None

    def reset(self):
        while self.is_reading(): time.sleep(0.01)
        if self._proc: self._proc.reset()

    def __del__(self):
        try:
            self.reset()
        except AttrbuteError:
            pass

    def configure(self,cfg, color=0):
        # if "this" was configured before, reset it
        if self._name: self.reset()

        # get name
        if not cfg['filler_name']:
            sys.stderr.write('filler_name is empty!\n')
            raise ValueError

        # ensure unique name
        if self.__class__.exist(cfg['filler_name']) and not self.__class__.instance_by_name(cfg['filler_name']) == self:
            sys.stderr.write('filler_name %s already running!' % cfg['filler_name'])
            return
        self._name = cfg['filler_name']

        # get QueueProcessor config file
        self._cfg_file = cfg['filler_cfg']
        if not self._cfg_file or not os.path.isfile(self._cfg_file):
            sys.stderr.write('filler_cfg file does not exist: %s\n' % self._cfg_file)
            raise ValueError

        # set verbosity
        if 'verbosity' in cfg:
            self._verbose = bool(cfg['verbosity'])

        # configure thread processor
        self._proc = larcv.QueueProcessor(self._name)

        self._proc.configure(self._cfg_file, color)

        # fetch batch filler info
        self._storage = {}
        for i in range(len(self._proc.batch_fillers())):
            pid = self._proc.batch_fillers()[i]
            name = self._proc.storage_name(pid)
            dtype = larcv.BatchDataTypeName(self._proc.batch_types()[i])
            self._storage[name]=batch_pydata(dtype)
            if 'make_copy' in cfg and cfg['make_copy']:
                self._storage[name]._make_copy = True

        # all success?
        # register *this* instance
        self.__class__._instance_m[self._name] = self

    def set_next_batch(self, batch_indexes):
        # if type(batch_indexes) != larcv.VectorOfSizet:
        #     indexes = larcv.VectorOfSizet()
        #     indexes.resize(len(batch_indexes))
        #     for i, val in enumerate(batch_indexes):
        #         indexes[i] = int(val)
        #     batch_indexes = indexes
        self._proc.set_next_batch(batch_indexes)

    def batch_process(self):
        while self.is_reading():
            time.sleep(0.01)
        self._proc.batch_process()


    def prepare_next(self):
        self._proc.prepare_next()

    def is_reading(self,storage_id=None):
        return self._proc.is_reading()

    def pop_current_data(self):
        # Promote the "next" data to current in C++ and release current
        self._proc.pop_current_data()

    def next(self,store_entries=False,store_event_ids=False):

        # Calling next will load the next set of data into batch_pydata.  It does not do any
        # automatic data loading or steping, you must do this manually.

        for name,storage in self._storage.items():
            dtype = storage.dtype()
            if dtype == "float32":
                factory = larcv.BatchDataQueueFactoryFloat.get()
            elif dtype == "float64":
                factory = larcv.BatchDataQueueFactoryDouble.get()
            elif dtype == "int":
                factory = larcv.BatchDataQueueFactoryInt.get()
            # These here below are NOT yet wrapped with swig.  Submit a ticket if you need them!
            # elif dtype == "char":
            #    factory = larcv.BatchDataQueueFactoryDouble.get()
            # elif dtype == "short":
            #    factory = larcv.BatchDataQueueFactoryDouble.get()
            # elif dtype == "string":
            #    factory = larcv.BatchDataQueueFactoryDouble.get()
            else:
                factory = None
            batch_storage = factory.get_queue(name)

            batch_data = factory.get_queue(name).get_batch()
            storage.set_data(storage_id=name, larcv_batchdata=batch_data)

        if not store_entries: self._event_entries = None
        else: self._event_entries = self._proc.processed_entries()

        if not store_event_ids: self._event_ids = None
        else: self._event_ids = self._proc.processed_events()

        return

    def fetch_data(self,key):
        try:
            return self._storage[key]
        except KeyError:
            sys.stderr.write('Cannot fetch data w/ key %s (unknown)\n' % key)
            return

    def fetch_event_ids(self):
        return self._event_ids

    def fetch_entries(self):
        return self._event_entries

    def fetch_n_entries(self):
        return self._proc.get_n_entries()



import pytest
import larcv
import numpy

from random import Random
random = Random()


@pytest.mark.parametrize('dimension', [1, 2, 3, 4])
def test_Tensor(dimension):

    # Create a numpy array of the specified dimension with random values:


    shape = [] 
    for dim in range(dimension):
        if dimension < 3:
            shape.append(random.randint(1, 1e3))
        else:
            shape.append(random.randint(1, 20))
                
    raw_image = numpy.random.random(shape).astype("float32")

    # Create a Tensor from the array:
    if dimension == 1:
        t = larcv.Tensor1D(raw_image)
    elif dimension == 2:
        t = larcv.Tensor2D(raw_image)
    elif dimension == 3:
        t = larcv.Tensor3D(raw_image)
    elif dimension == 4:
        t = larcv.Tensor4D(raw_image)

    assert t.size() == numpy.prod(shape)

    for i in range(20):
        index = random.randint(0,t.size() - 1)
        assert t.pixel(index) == raw_image.take(index)

    nd_t = t.as_array()

    assert (nd_t - raw_image).sum() < 1e-6

    # Verify the tensor and array match by checking elements:
    return True

@pytest.mark.parametrize('dimension', [1, 2, 3, 4])
@pytest.mark.parametrize('pooling', [larcv.kPoolAverage, larcv.kPoolMax, larcv.kPoolSum])
def test_tensor_compression(dimension, pooling):
    
    try:
        from skimage.measure import block_reduce
    except:
        pytest.skip("Could not import skimage")

    shape = [] 
    for dim in range(dimension):
        shape.append(64)
                
    raw_image = numpy.random.random(shape).astype("float32")
    # Create a Tensor from the array:
    if dimension == 1:
        t = larcv.Tensor1D(raw_image)
    elif dimension == 2:
        t = larcv.Tensor2D(raw_image)
    elif dimension == 3:
        t = larcv.Tensor3D(raw_image)
    elif dimension == 4:
        t = larcv.Tensor4D(raw_image)

    # Pooling in numpy is not trivial.  So instead of checking every element,
    # We will check against random elements.

    compression = 2

    t_compressed = t.compress(compression, pooling)

    compressed_numpy = t_compressed.as_array()

    # Use skimage to compress the original:
    kernel = (compression,) * dimension
    if pooling == larcv.kPoolAverage:
        sk_reduced = block_reduce(raw_image, kernel,func=numpy.mean)
    elif pooling == larcv.kPoolMax:
        sk_reduced = block_reduce(raw_image, kernel,func=numpy.max)
    elif pooling == larcv.kPoolSum:
        sk_reduced = block_reduce(raw_image, kernel,func=numpy.sum)

    # Do some checks:
    assert numpy.abs( (compressed_numpy.sum() - sk_reduced.sum()) / sk_reduced.sum() )< 1e-6
    max_index = numpy.prod(compressed_numpy.shape)
    for i in range(50):
        index = numpy.random.randint(0, max_index)
        assert numpy.abs(compressed_numpy.take(index) - sk_reduced.take(index) )< 1e-4


if __name__ == '__main__':
    test_Tensor(1)
    test_Tensor(2)
    test_Tensor(3)
    test_Tensor(4)



import pytest
import unittest
import larcv

from random import Random
random = Random()

from larcv import data_generator

@pytest.fixture()
def rand_num_events():
    return random.randint(5, 100)



@pytest.mark.parametrize('dimension', [2, 3])
@pytest.mark.parametrize('n_projections', [1, 2, 3])
def test_write_sparse_tensors(tmpdir, rand_num_events, dimension, n_projections):

    voxel_set_list = data_generator.build_sparse_tensor(rand_num_events, n_projections = n_projections)

    random_file_name = str(tmpdir + "/test_write_sparse_tensors.h5")

    data_generator.write_sparse_tensors(random_file_name, voxel_set_list, dimension, n_projections)


@pytest.mark.parametrize('dimension', [2, 3])
@pytest.mark.parametrize('n_projections', [1, 2, 3])
def test_read_write_sparse_tensors(tmpdir, rand_num_events, dimension, n_projections):

    import numpy

    random_file_name = str(tmpdir + "/test_write_read_sparse_tensors.h5")

    voxel_set_list = data_generator.build_sparse_tensor(rand_num_events, n_projections = n_projections)

    data_generator.write_sparse_tensors(random_file_name, voxel_set_list, dimension, n_projections)
    read_voxel_set_list = data_generator.read_sparse_tensors(random_file_name, dimension)

    # Check the same number of events came back:
    assert(len(read_voxel_set_list) == rand_num_events)
    for event in range(rand_num_events):
        # Check the same number of projections per event:
        assert(len(read_voxel_set_list[event]) == len(voxel_set_list[event]))
        for projection in range(n_projections):
            # Check the same number of voxels:
            input_voxelset = voxel_set_list[event][projection]
            read_voxelset = read_voxel_set_list[event][projection]
            assert(read_voxelset['n_voxels'] == input_voxelset['n_voxels'])

            # Check voxel properties:
            # Sum of indexes
            # Sum of values
            # std of values
            if input_voxelset['n_voxels'] == 0:
                continue
            print(input_voxelset['values'])
            assert(numpy.sum(input_voxelset['indexes']) == numpy.sum(read_voxelset['indexes']))
            assert( abs( numpy.sum(input_voxelset['values']) - numpy.sum(read_voxelset['values']) ) < 1e-3 )
            assert( abs( numpy.std(input_voxelset['values']) - numpy.std(read_voxelset['values']) ) < 1e-3 )



if __name__ == '__main__':
    tmpdir = "./"
    rand_num_events = 5
    n_projections = 3
    dimension = 2
    test_read_write_sparse_tensors(tmpdir, rand_num_events, dimension, n_projections)








